__all__ = [
    'ubcMeshReaderBase',
    'ModelAppenderBase',
]

__displayname__ = 'Base Classes'

# Outside Imports:
import os

import numpy as np
import vtk

from .. import _helpers, base

###############################################################################


# UBC Mesh Reader Base
class ubcMeshReaderBase(base.TwoFileReaderBase):
    """A base class for the UBC mesh readers
    """
    __displayname__ = 'UBC Mesh Reader Base'
    __category__ = 'base'
    extensions = 'mesh msh dat txt text'

    def __init__(self, nOutputPorts=1, outputType='vtkUnstructuredGrid', **kwargs):
        base.TwoFileReaderBase.__init__(self,
                                        nOutputPorts=nOutputPorts, outputType=outputType,
                                        **kwargs)
        self.__data_name = 'Data'
        self.__use_filename = True # flag on whether or not to use the model file
        # extension as data name
        # For keeping track of type (2D vs 3D)
        self.__sizeM = None


    def is_3d(self):
        """Returns true if mesh is spatially references in three dimensions"""
        return self.__sizeM.shape[0] >= 3

    def is_2d(self):
        """Returns true if mesh is spatially references in only two dimensions"""
        return self.__sizeM.shape[0] == 1

    @staticmethod
    def _ubc_mesh_2d_part(FileName):
        """Internal helper to read 2D mesh file"""
        # This is a helper method to read file contents of mesh
        try:
            fileLines = np.genfromtxt(FileName, dtype=str, delimiter='\n', comments='!')
        except (IOError, OSError) as fe:
            raise _helpers.PVGeoError(str(fe))

        def _genTup(sft, n):
            # This reads in the data for a dimension
            pts = []
            disc = []
            for i in range(n):
                ln = fileLines[i+sft].split('!')[0].split()
                if i == 0:
                    o = ln[0]
                    pts.append(o)
                    ln = [ln[1],ln[2]]
                pts.append(ln[0])
                disc.append(ln[1])
            return pts, disc

        # Get the number of lines for each dimension
        nx = int(fileLines[0].split('!')[0])
        nz = int(fileLines[nx+1].split('!')[0])

        # Get the origins and tups for both dimensions
        xpts, xdisc = _genTup(1, nx)
        zpts, zdisc = _genTup(2+nx, nz)

        return xpts, xdisc, zpts, zdisc

    def _read_extent(self):
        """Reads the mesh file for the UBC 2D/3D Mesh or OcTree format to get
        output extents. Computationally inexpensive method to discover whole
        output extent.

        Return:
            tuple(int) :
                This returns a tuple of the whole extent for the grid to be
                made of the input mesh file (0,n1-1, 0,n2-1, 0,n3-1). This
                output should be directly passed to set the whole output extent.

        """
        # Read the mesh file as line strings, remove lines with comment = !
        v = np.array(np.__version__.split('.')[0:2], dtype=int)
        FileName = self.get_mesh_filename()
        try:
            if v[0] >= 1 and v[1] >= 10:
                # max_rows in numpy versions >= 1.10
                msh = np.genfromtxt(FileName, delimiter='\n', dtype=np.str, comments='!', max_rows=1)
            else:
                # This reads whole file :(
                msh = np.genfromtxt(FileName, delimiter='\n', dtype=np.str, comments='!')[0]
        except (IOError, OSError) as fe:
            raise _helpers.PVGeoError(str(fe))
        # Fist line is the size of the model
        self.__sizeM = np.array(msh.ravel()[0].split(), dtype=int)
        # Check if the mesh is a UBC 2D mesh
        if self.__sizeM.shape[0] == 1:
            # Read in data from file
            xpts, xdisc, zpts, zdisc = ubcMeshReaderBase._ubc_mesh_2d_part(FileName)
            nx = np.sum(np.array(xdisc,dtype=int))+1
            nz = np.sum(np.array(zdisc,dtype=int))+1
            return (0,nx, 0,1, 0,nz)
        # Check if the mesh is a UBC 3D mesh or OcTree
        elif self.__sizeM.shape[0] >= 3:
            # Get mesh dimensions
            dim = self.__sizeM[0:3]
            ne,nn,nz = dim[0], dim[1], dim[2]
            return (0,ne, 0,nn, 0,nz)
        else:
            raise _helpers.PVGeoError('File format not recognized')


    @staticmethod
    def ubc_model_3d(FileName):
        """Reads the 3D model file and returns a 1D NumPy float array. Use the
        place_model_on_mesh() method to associate with a grid.

        Args:
            FileName (str) : The model file name(s) as an absolute path for the
                input model file in UBC 3D Model Model Format. Also accepts a
                `list` of string file names.

        Return:
            np.array :
                Returns a NumPy float array that holds the model data
                read from the file. Use the ``place_model_on_mesh()`` method to
                associate with a grid. If a list of file names is given then it
                will return a dictionary of NumPy float array with keys as the
                basenames of the files.
        """
        # Check if recurssion needed
        if isinstance(FileName, (list, tuple)):
            out = {}
            for f in FileName:
                out[os.path.basename(f)] = ubcMeshReaderBase.ubc_model_3d(f)
            return out
        # Perform IO
        try:
            data = np.genfromtxt(FileName, dtype=np.float, comments='!')
        except (IOError, OSError) as fe:
            raise _helpers.PVGeoError(str(fe))
        return data

    def set_use_filename(self, flag):
        """Set a flag on whether or not to use the filename as the data array name"""
        if self.__use_filename != flag:
            self.__use_filename = flag
            self.Modified(read_again_mesh=False, read_again_models=False)

    def set_data_name(self, name):
        """Set the data array name"""
        if name == '':
            self.__use_filename = True
            self.Modified(read_again_mesh=False, read_again_models=False)
        elif self.__data_name != name:
            self.__data_name = name
            self.__use_filename = False
            self.Modified(read_again_mesh=False, read_again_models=False)

    def get_data_name(self):
        """Get the data array name"""
        if self.__use_filename:
            mname = self.get_model_filenames(idx=0)
            return os.path.basename(mname)
        return self.__data_name




###############################################################################


# UBC Model Appender Base
class ModelAppenderBase(base.AlgorithmBase):
    """A base class for create mesh-model appenders on the UBC Mesh formats
    """
    __displayname__ = 'Model Appender Base'
    __category__ = 'base'

    def __init__(self, inputType='vtkRectilinearGrid', outputType='vtkRectilinearGrid', **kwargs):
        base.AlgorithmBase.__init__(self,
                                    nInputPorts=1, inputType=inputType,
                                    nOutputPorts=1, outputType=outputType)
        self._model_filenames = kwargs.get('model_files', [])
        self.__data_name = kwargs.get('dataname', 'Appended Data')
        self.__use_filename = True
        self._models = []
        self.__need_to_read = True
        self._is_3D = None
        # For the VTK/ParaView pipeline
        self.__dt = kwargs.get('dt', 1.0)
        self.__timesteps = None
        self.__last_successfull_index = 0 #This is the index to use if the current timestep is unavailable


    def need_to_read(self, flag=None):
        """Ask self if the reader needs to read the files again

        Args:
            flag (bool): if the flag is set then this method will set the read
                status

        Return:
            bool:
                The status of the reader aspect of the filter.
        """
        if flag is not None and isinstance(flag, (bool, int)):
            self.__need_to_read = flag
            self._update_time_steps()
        return self.__need_to_read

    def Modified(self, read_again=True):
        """Call modified if the files needs to be read again again.
        """
        if read_again:
            self.__need_to_read = read_again
        base.AlgorithmBase.Modified(self)

    def modified(self, read_again=True):
        """Call modified if the files needs to be read again again.
        """
        return self.Modified(read_again=read_again)

    def _update_time_steps(self):
        """For internal use only: appropriately sets the timesteps.
        """
        # Use the inputs' timesteps: this merges the timesteps values
        ts0 = _helpers.get_input_time_steps(self, port=0)
        if ts0 is None:
            ts0 = np.array([])
        ts1 = _helpers._calculate_time_range(len(self._model_filenames), self.__dt)
        tsAll = np.unique(np.concatenate((ts0, ts1), 0))
        # Use both inputs' time steps
        self.__timesteps = _helpers.update_time_steps(self, tsAll, explicit=True)
        return 1

    def _read_up_front(self):
        raise NotImplementedError()

    def _place_on_mesh(self, output, idx=0):
        raise NotImplementedError()


    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to generate output
        """
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        output = self.GetOutputData(outInfo, 0)
        output.DeepCopy(pdi) # ShallowCopy if you want changes to propagate upstream
        # Get requested time index
        i = _helpers.get_requested_time(self, outInfo)
        # Perfrom task:
        if self.__need_to_read:
            self._read_up_front()
        # Place the model data for given timestep onto the mesh
        if len(self._models) > i:
            self._place_on_mesh(output, idx=i)
            self.__last_successfull_index = i
        else:
            # put the last array as a placeholder
            self._place_on_mesh(output, idx=self.__last_successfull_index)
        return 1

    def RequestInformation(self, request, inInfo, outInfo):
        """Used by pipeline to handle time variance and update output extents
        """
        self._update_time_steps()
        pdi = self.GetInputData(inInfo, 0, 0)
        # Determine if 2D or 3D and read
        if isinstance(pdi, vtk.vtkRectilinearGrid) and pdi.GetExtent()[3] == 1:
            self._is_3D = False
        else:
            self._is_3D = True
        return 1

    #### Setters and Getters ####

    def has_models(self):
        """Return True if models are associated with this mesh"""
        return len(self._model_filenames) > 0

    def get_time_step_values(self):
        """Use this in ParaView decorator to register timesteps.
        """
        # if unset, force at least one attempt to set the timesteps
        if self.__timesteps is None:
            self._update_time_steps()
        return self.__timesteps if self.__timesteps is not None else None

    def clear_models(self):
        """Use to clear data file names.
        """
        self._model_filenames = []
        self._models = []
        self.Modified(read_again=True)

    def add_model_file_name(self, filename):
        """Use to set the file names for the reader. Handles single string or
        list of strings.
        """
        if filename is None:
            return # do nothing if None is passed by a constructor on accident
        elif isinstance(filename, (list, tuple)):
            for f in filename:
                self.add_model_file_name(f)
            self.Modified()
        elif filename not in self._model_filenames:
            self._model_filenames.append(filename)
            self.Modified()
        return 1

    def get_model_filenames(self, idx=None):
        """Returns the list of file names or given and index returns a specified
        timestep's filename.
        """
        if idx is None or not self.has_models():
            return self._model_filenames
        return self._model_filenames[idx]

    def set_use_filename(self, flag):
        """Set a flag on whether or not to use the filename as the data array name"""
        if self.__use_filename != flag:
            self.__use_filename = flag
            self.Modified(read_again=False)

    def set_data_name(self, name):
        """Set the data array name"""
        if name == '':
            self.__use_filename = True
            self.Modified(read_again=False)
        elif self.__data_name != name:
            self.__data_name = name
            self.__use_filename = False
            self.Modified(read_again=False)

    def get_data_name(self):
        """Get the data array name"""
        if self.__use_filename:
            mname = self.get_model_filenames(idx=0)
            return os.path.basename(mname)
        return self.__data_name



__all__ = [
    '_calculate_time_range',
    'update_time_steps',
    'get_requested_time',
    'get_input_time_steps',
    'get_combined_input_time_steps',
]

import collections

import numpy as np


def _calculate_time_range(nt, dt=1.0):
    """Discretizes time range accoridng to step size ``dt`` in seconds
    """
    return np.arange(0,nt*dt,dt, dtype=float)


def update_time_steps(algorithm, nt, dt=1.0, explicit=False):
    """Handles setting up the timesteps on on the pipeline for a file series reader.

    Args:
        algorithm (vtkDataObject): The data object (Proxy) on the pipeline
            (pass `self` from algorithm subclasses)
        nt (int or list): Number of timesteps (Pass a list to use length of
            that list)
        dt (float): The discrete value in seconds for the time step.
        explicit (boolean): if true, this will treat the nt argument as the exact
            timestep values to use

    Return:
        numpy.array : Returns the timesteps as an array
    """
    if explicit and isinstance(nt, collections.Iterable):
        timesteps = nt
    else:
        if isinstance(nt, collections.Iterable):
            nt = len(nt)
        timesteps = _calculate_time_range(nt, dt=1.0)
    if len(timesteps) < 1:
        # NOTE: we may want to raise a warning here on the dev side.
        #       if developing a new algorithm that uses this, you may want to
        #       know exactly when this failse to update
        #'update_time_steps() is not updating because passed time step values are NULL.'
        return None
    executive = algorithm.GetExecutive()
    oi = executive.GetOutputInformation(0)
    #oi = outInfo.GetInformationObject(0)
    oi.Remove(executive.TIME_STEPS())
    oi.Remove(executive.TIME_RANGE())
    for t in timesteps:
        oi.Append(executive.TIME_STEPS(), t)
    oi.Append(executive.TIME_RANGE(), timesteps[0])
    oi.Append(executive.TIME_RANGE(), timesteps[-1])
    return timesteps


def get_requested_time(algorithm, outInfo, idx=0):
    """Handles setting up the timesteps on on the pipeline for a file series
    reader.

    Args:
        algorithm (vtkDataObject) : The data object (Proxy) on the pipeline
            (pass `self` from algorithm subclasses)
        outInfo (vtkInformationVector) : The output information for the
            algorithm
        idx (int) : the index for the output port

    Return:
        int : the index of the requested time

    Example:
        >>> # Get requested time index
        >>> i = _helpers.get_requested_time(self, outInfo)
    """
    executive = algorithm.GetExecutive()
    timesteps = algorithm.get_time_step_values()
    outInfo = outInfo.GetInformationObject(idx)
    if timesteps is None or len(timesteps) == 0:
        return 0
    elif outInfo.Has(executive.UPDATE_TIME_STEP()) and len(timesteps) > 0:
        utime = outInfo.Get(executive.UPDATE_TIME_STEP())
        return np.argmin(np.abs(np.array(timesteps) - utime))
    else:
        # if we cant match the time, give first
        if not len(timesteps) > 0:
            raise AssertionError('Number of timesteps must be greater than 0')
        return 0


def get_input_time_steps(algorithm, port=0, idx=0):
    """Get the timestep values for the algorithm's input

    Args:
        algorithm (vtkDataObject) : The data object (Proxy) on the pipeline
            (pass `self` from algorithm subclasses)
        port (int)  : the input port
        idx (int) : optional : the connection index on the input port

    Return:
        list : the time step values of the input (if there arn't any, returns ``None``)
    """
    executive = algorithm.GetExecutive()
    ii = executive.GetInputInformation(port, idx)
    return ii.Get(executive.TIME_STEPS())


def get_combined_input_time_steps(algorithm, idx=0):
    """This will iterate over all input ports and combine their unique timesteps
    for an output algorithm to have.

    Args:
        algorithm (vtkDataObject) : The data object (Proxy) on the pipeline
            (pass `self` from algorithm subclasses)

    Return:
        np.ndarray : a 1D array of all the unique timestep values (empty array if no time variance)
    """
    executive = algorithm.GetExecutive()
    tsteps = []
    for port in range(executive.GetNumberOfInputPorts()):
        ii = executive.GetInputInformation(port, idx)
        ti = ii.Get(executive.TIME_STEPS())
        if ti is None:
            ti = np.array([])
        tsteps.append(ti)
    return np.unique(np.concatenate(tsteps, 0))



__all__ = [
    'OMFReader',
]

import omf
import omfvista
import vtk

from .. import _helpers
from ..base import ReaderBaseBase


class OMFReader(ReaderBaseBase):
    """Handles reading an OMF Project
    """
    __displayname__ = 'OMF Project Reader'
    __category__ = 'reader'
    extensions = 'omf'
    description = 'PVGeo: Open Mining Format Project'

    def __init__(self):
        ReaderBaseBase.__init__(self,
                                nOutputPorts=1, outputType='vtkMultiBlockDataSet')
        # Properties:
        self._dataselection = vtk.vtkDataArraySelection()
        self._dataselection.AddObserver("ModifiedEvent", _helpers.create_modified_callback(self))
        self.__names = []
        self.__data = dict()
        self.__project = None

    def Modified(self, read_again=False):
        """Ensure default is overridden to be false so array selector can call.
        """
        ReaderBaseBase.Modified(self, read_again=read_again)

    def modified(self, read_again=False):
        """Ensure default is overridden to be false so array selector can call.
        """
        return self.Modified(read_again=read_again)

    def get_file_name(self):
        """Super class has file names as a list but we will only handle a single
        project file. This provides a conveinant way of making sure we only
        access that single file.
        A user could still access the list of file names using ``get_file_names()``.
        """
        return ReaderBaseBase.get_file_names(self, idx=0)

    #### Methods for performing the read ####

    def _read_up_front(self):
        """Internal functiona to read all data at the start"""
        # Read all elements
        reader = omf.OMFReader(self.get_file_name())
        self.__project = reader.get_project()
        self.__names = [e.name for e in self.__project.elements]
        for n in self.__names:
            self._dataselection.AddArray(n)
        self.need_to_read(flag=False)
        return 1

    def _get_raw_data(self):
        """Converts OMF data to VTK data objects."""
        # Now iterate over the elements and add converted data to the data dict:
        data = dict()
        for e in self.__project.elements:
            if self._dataselection.ArrayIsEnabled(e.name):
                if e.name not in self.__data:
                    self.__data[e.name] = omfvista.wrap(e)
                data[e.name] = self.__data[e.name]
        return data

    #### pipeline methods ####

    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to get data for current timestep and populate the output data object.
        """
        # Get output:
        #output = self.GetOutputData(outInfo, 0)
        output = vtk.vtkMultiBlockDataSet.GetData(outInfo, 0)
        # Perfrom the read
        if self.need_to_read():
            self._read_up_front()
        data = self._get_raw_data()
        # Set number of blocks based on user choice in the selction
        output.SetNumberOfBlocks(self._dataselection.GetNumberOfArraysEnabled())
        blk = 0
        # iterate over data set to produce output based on users selection
        keys = data.keys()
        for name in keys:
            output.SetBlock(blk, data[name])
            output.GetMetaData(blk).Set(vtk.vtkCompositeDataSet.NAME(), name)
            blk += 1
        return 1


    #### Getters / Setters ####


    def GetDataSelection(self):
        """Return the current user selection of data elements"""
        if self.need_to_read():
            self._read_up_front()
        return self._dataselection



paraview_plugin_version = '2.1.0'
# This is module to import. It provides VTKPythonAlgorithmBase, the base class
# for all python-based vtkAlgorithm subclasses in VTK and decorators used to
# 'register' the algorithm with ParaView along with information about UI.
from paraview.util.vtkAlgorithm import smproperty, smproxy

# Helpers:
from PVGeo import _helpers
# Classes to Decorate
from PVGeo.readers import PackedBinariesReader, MadagascarReader
from PVGeo.readers import DelimitedTextReader, XYZTextReader


###############################################################################



@smproxy.reader(name="PVGeoPackedBinariesReader",
                label='PVGeo: %s' % PackedBinariesReader.__displayname__,
                extensions=PackedBinariesReader.extensions,
                file_description=PackedBinariesReader.description)
class PVGeoPackedBinariesReader(PackedBinariesReader):
    def __init__(self):
        PackedBinariesReader.__init__(self)

    #### Seters and Geters ####
    @smproperty.xml(_helpers.get_file_reader_xml(PackedBinariesReader.extensions, reader_description=PackedBinariesReader.description))
    def AddFileName(self, filename):
        PackedBinariesReader.AddFileName(self, filename)



    @smproperty.doublevector(name="TimeDelta", default_values=1.0, panel_visibility="advanced")
    def set_time_delta(self, dt):
        PackedBinariesReader.set_time_delta(self, dt)

    @smproperty.doublevector(name="TimestepValues", information_only="1", si_class="vtkSITimeStepsProperty")
    def get_time_step_values(self):
        """This is critical for registering the timesteps"""
        return PackedBinariesReader.get_time_step_values(self)

    @smproperty.xml(_helpers.get_drop_down_xml('Endian','set_endian',
                                               ['Native', 'Little-Endian', 'Big-Endian'],
                                               help='This is the type memory endianness.'))
    def set_endian(self, endian):
        PackedBinariesReader.set_endian(self, endian)

    @smproperty.xml(_helpers.get_drop_down_xml('DataType','set_data_type',
                                               ['Float64', 'Float32', 'Integer4'],
                                               help='This is data type to read.'))
    def set_data_type(self, dtype):
        PackedBinariesReader.set_data_type(self, dtype)


    @smproperty.stringvector(name='DataName', default_values='Data')
    def set_data_name(self, data_name):
        PackedBinariesReader.set_data_name(self, data_name)


###############################################################################



@smproxy.reader(name="PVGeoMadagascarReader",
                label='PVGeo: %s' % MadagascarReader.__displayname__,
                extensions=MadagascarReader.extensions,
                file_description=MadagascarReader.description)
class PVGeoMadagascarReader(MadagascarReader):
    def __init__(self):
        MadagascarReader.__init__(self)

    #### Seters and Geters ####


    @smproperty.xml(_helpers.get_file_reader_xml(MadagascarReader.extensions, reader_description=MadagascarReader.description))
    def AddFileName(self, filename):
        MadagascarReader.AddFileName(self, filename)


    @smproperty.doublevector(name="TimeDelta", default_values=1.0, panel_visibility="advanced")
    def set_time_delta(self, dt):
        MadagascarReader.set_time_delta(self, dt)

    @smproperty.doublevector(name="TimestepValues", information_only="1", si_class="vtkSITimeStepsProperty")
    def get_time_step_values(self):
        """This is critical for registering the timesteps"""
        return MadagascarReader.get_time_step_values(self)

    @smproperty.xml(_helpers.get_drop_down_xml('Endian','set_endian',
                                               ['Native', 'Little-Endian', 'Big-Endian'],
                                               help='This is the type memory endianness.'))
    def set_endian(self, endian):
        MadagascarReader.set_endian(self, endian)

    @smproperty.xml(_helpers.get_drop_down_xml('DataType','set_data_type',
                                               ['Float64', 'Float32', 'Integer4'],
                                               help='This is data type to read.'))
    def set_data_type(self, dtype):
        MadagascarReader.set_data_type(self, dtype)


    @smproperty.stringvector(name='DataName', default_values='Data')
    def set_data_name(self, data_name):
        MadagascarReader.set_data_name(self, data_name)



###############################################################################



@smproxy.reader(name="PVGeoDelimitedTextReader",
                label='PVGeo: %s' % DelimitedTextReader.__displayname__,
                extensions=DelimitedTextReader.extensions,
                file_description=DelimitedTextReader.description)
class PVGeoDelimitedTextReader(DelimitedTextReader):
    def __init__(self):
        DelimitedTextReader.__init__(self)


    #### Seters and Geters ####
    @smproperty.xml(_helpers.get_file_reader_xml(DelimitedTextReader.extensions, reader_description=DelimitedTextReader.description))
    def AddFileName(self, filename):
        DelimitedTextReader.AddFileName(self, filename)



    @smproperty.doublevector(name="TimeDelta", default_values=1.0, panel_visibility="advanced")
    def set_time_delta(self, dt):
        DelimitedTextReader.set_time_delta(self, dt)

    @smproperty.doublevector(name="TimestepValues", information_only="1", si_class="vtkSITimeStepsProperty")
    def get_time_step_values(self):
        """This is critical for registering the timesteps"""
        return DelimitedTextReader.get_time_step_values(self)

    @smproperty.stringvector(name="Delimiter", default_values=" ")
    def set_delimiter(self, deli):
        DelimitedTextReader.set_delimiter(self, deli)

    @smproperty.xml(_helpers.get_property_xml(name='Use Split on Whitespace', command='set_split_on_white_space', default_values=False, help='A boolean to override the Delimiter_Field and use whitespace as delimiter.'))
    def set_split_on_white_space(self, flag):
        DelimitedTextReader.set_split_on_white_space(self, flag)

    @smproperty.intvector(name="SkipRows", default_values=0)
    def set_skip_rows(self, skip):
        DelimitedTextReader.set_skip_rows(self, skip)

    @smproperty.stringvector(name="Comments", default_values="!")
    def set_comments(self, identifier):
        DelimitedTextReader.set_comments(self, identifier)

    @smproperty.xml(_helpers.get_property_xml(name='Has Titles', command='set_has_titles', default_values=False, help='A boolean for if the delimited file has header titles for the data arrays.'))
    def set_has_titles(self, flag):
        DelimitedTextReader.set_has_titles(self, flag)



###############################################################################



@smproxy.reader(name="PVGeoXYZTextReader",
                label='PVGeo: %s' % XYZTextReader.__displayname__,
                extensions=XYZTextReader.extensions,
                file_description=XYZTextReader.description)
class PVGeoXYZTextReader(XYZTextReader):
    def __init__(self):
        XYZTextReader.__init__(self)


    #### Seters and Geters ####
    @smproperty.xml(_helpers.get_file_reader_xml(XYZTextReader.extensions, reader_description=XYZTextReader.description))
    def AddFileName(self, filename):
        XYZTextReader.AddFileName(self, filename)



    @smproperty.doublevector(name="TimeDelta", default_values=1.0, panel_visibility="advanced")
    def set_time_delta(self, dt):
        XYZTextReader.set_time_delta(self, dt)

    @smproperty.doublevector(name="TimestepValues", information_only="1", si_class="vtkSITimeStepsProperty")
    def get_time_step_values(self):
        """This is critical for registering the timesteps"""
        return XYZTextReader.get_time_step_values(self)

    @smproperty.stringvector(name="Delimiter", default_values=" ")
    def set_delimiter(self, deli):
        XYZTextReader.set_delimiter(self, deli)

    @smproperty.xml(_helpers.get_property_xml(name='Use Split on Whitespace', command='set_split_on_white_space', default_values=False, help='A boolean to override the Delimiter_Field and use whitespace as delimiter.'))
    def set_split_on_white_space(self, flag):
        XYZTextReader.set_split_on_white_space(self, flag)

    @smproperty.intvector(name="SkipRows", default_values=0)
    def set_skip_rows(self, skip):
        XYZTextReader.set_skip_rows(self, skip)

    @smproperty.stringvector(name="Comments", default_values="!")
    def set_comments(self, identifier):
        XYZTextReader.set_comments(self, identifier)



"""
These are helpers specifically for the file readers for private use only.
@author: Bane Sullivan
"""

__all__ = [
    'clean_data_name',
    'create_modified_callback',
]

import os


def clean_data_name(data_name, filename):
    """A helper to clean a filename to make a useful data array name"""
    if data_name is None or data_name == '':
        data_name = os.path.splitext(os.path.basename(filename))[0]
    return data_name


def create_modified_callback(anobject):
    import weakref
    weakref_obj = weakref.ref(anobject)
    anobject = None

    def _mark_modified(*args, **kwars):
        o = weakref_obj()
        if o is not None:
            o.Modified()

    return _mark_modified



paraview_plugin_version = '2.1.0'
# This is module to import. It provides VTKPythonAlgorithmBase, the base class
# for all python-based vtkAlgorithm subclasses in VTK and decorators used to
# 'register' the algorithm with ParaView along with information about UI.
from paraview.util.vtkAlgorithm import smhint, smproxy

from pyvista import examples
import vtk
from vtk.util.vtkAlgorithm import VTKPythonAlgorithmBase


MENU_CAT = 'PyVista: Example Data Sets'


class _ExampleLoader(VTKPythonAlgorithmBase):

    def __init__(self):
        VTKPythonAlgorithmBase.__init__(self, nInputPorts=0, nOutputPorts=1,
                                        outputType=self._example_data.GetClassName())


    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to get data for current timestep and populate the
        output data object.
        """
        # Get output:
        output = self.GetOutputData(outInfo, 0)
        output.DeepCopy(self._example_data)
        return 1


    def RequestInformation(self, request, inInfo, outInfo):
        """Used by pipeline to handle output extents"""
        try:
            extent = self._example_data.GetExtent()
            # Now set whole output extent
            # ext = [0, extent[0]-1, 0,extent[1]-1, 0,extent[2]-1]
            info = outInfo.GetInformationObject(0)
            # Set WHOLE_EXTENT: This is absolutely necessary
            info.Set(vtk.vtkStreamingDemandDrivenPipeline.WHOLE_EXTENT(), extent, 6)
        except AttributeError:
            pass
        return 1



@smproxy.source(name='PyVistaStHelens', label='Mt. St. Helens')
@smhint.xml('''<ShowInMenu category="%s"/>
    <RepresentationType view="RenderView" type="Surface" />''' % MENU_CAT)
class PyVistaStHelens(_ExampleLoader):
    def __init__(self):
        self._example_data = examples.download_st_helens()
        _ExampleLoader.__init__(self)


@smproxy.source(name='PyVistaChannels', label='Channels')
@smhint.xml('''<ShowInMenu category="%s"/>
    <RepresentationType view="RenderView" type="Surface" />''' % MENU_CAT)
class PyVistaChannels(_ExampleLoader):
    def __init__(self):
        self._example_data = examples.load_channels()
        _ExampleLoader.__init__(self)


@smproxy.source(name='PyVistaActionFigure', label='Action Figure')
@smhint.xml('''<ShowInMenu category="%s"/>
    <RepresentationType view="RenderView" type="Surface" />''' % MENU_CAT)
class PyVistaActionFigure(_ExampleLoader):
    def __init__(self):
        self._example_data = examples.download_action_figure()
        _ExampleLoader.__init__(self)


@smproxy.source(name='PyVistaArmadillo', label='Armadillo')
@smhint.xml('''<ShowInMenu category="%s"/>
    <RepresentationType view="RenderView" type="Surface" />''' % MENU_CAT)
class PyVistaArmadillo(_ExampleLoader):
    def __init__(self):
        self._example_data = examples.download_armadillo()
        _ExampleLoader.__init__(self)


@smproxy.source(name='PyVistaBloodVessels', label='Blood Vessels')
@smhint.xml('''<ShowInMenu category="%s"/>
    <RepresentationType view="RenderView" type="Surface" />''' % MENU_CAT)
class PyVistaBloodVessels(_ExampleLoader):
    def __init__(self):
        self._example_data = examples.download_blood_vessels()
        _ExampleLoader.__init__(self)



@smproxy.source(name='PyVistaBlow', label='Blow')
@smhint.xml('''<ShowInMenu category="%s"/>
    <RepresentationType view="RenderView" type="Surface" />''' % MENU_CAT)
class PyVistaBlow(_ExampleLoader):
    def __init__(self):
        self._example_data = examples.download_blow()
        _ExampleLoader.__init__(self)


@smproxy.source(name='PyVistaBoltNut', label='Bolt & Nut')
@smhint.xml('''<ShowInMenu category="%s"/>
    <RepresentationType view="RenderView" type="Surface" />''' % MENU_CAT)
class PyVistaBoltNut(_ExampleLoader):
    def __init__(self):
        self._example_data = examples.download_bolt_nut()
        _ExampleLoader.__init__(self)


@smproxy.source(name='PyVistaBrain', label='Brain')
@smhint.xml('''<ShowInMenu category="%s"/>
    <RepresentationType view="RenderView" type="Surface" />''' % MENU_CAT)
class PyVistaBrain(_ExampleLoader):
    def __init__(self):
        self._example_data = examples.download_brain()
        _ExampleLoader.__init__(self)


@smproxy.source(name='PyVistaBunny', label='Bunny')
@smhint.xml('''<ShowInMenu category="%s"/>
    <RepresentationType view="RenderView" type="Surface" />''' % MENU_CAT)
class PyVistaBunny(_ExampleLoader):
    def __init__(self):
        self._example_data = examples.download_bunny()
        _ExampleLoader.__init__(self)


@smproxy.source(name='PyVistaCadModel', label='Cad Model')
@smhint.xml('''<ShowInMenu category="%s"/>
    <RepresentationType view="RenderView" type="Surface" />''' % MENU_CAT)
class PyVistaCadModel(_ExampleLoader):
    def __init__(self):
        self._example_data = examples.download_cad_model()
        _ExampleLoader.__init__(self)



@smproxy.source(name='PyVistaCarotid', label='Carotid')
@smhint.xml('''<ShowInMenu category="%s"/>
    <RepresentationType view="RenderView" type="Surface" />''' % MENU_CAT)
class PyVistaCarotid(_ExampleLoader):
    def __init__(self):
        self._example_data = examples.download_carotid()
        _ExampleLoader.__init__(self)


@smproxy.source(name='PyVistaCow', label='Cow')
@smhint.xml('''<ShowInMenu category="%s"/>
    <RepresentationType view="RenderView" type="Surface" />''' % MENU_CAT)
class PyVistaCow(_ExampleLoader):
    def __init__(self):
        self._example_data = examples.download_cow()
        _ExampleLoader.__init__(self)


@smproxy.source(name='PyVistaCowHead', label='Cow Head')
@smhint.xml('''<ShowInMenu category="%s"/>
    <RepresentationType view="RenderView" type="Surface" />''' % MENU_CAT)
class PyVistaCowHead(_ExampleLoader):
    def __init__(self):
        self._example_data = examples.download_cow_head()
        _ExampleLoader.__init__(self)


@smproxy.source(name='PyVistaNefertiti', label='Nefertiti')
@smhint.xml('''<ShowInMenu category="%s"/>
    <RepresentationType view="RenderView" type="Surface" />''' % MENU_CAT)
class PyVistaNefertiti(_ExampleLoader):
    def __init__(self):
        self._example_data = examples.download_nefertiti()
        _ExampleLoader.__init__(self)


@smproxy.source(name='PyVistaLidar', label='Lidar')
@smhint.xml('''<ShowInMenu category="%s"/>
    <RepresentationType view="RenderView" type="Surface" />''' % MENU_CAT)
class PyVistaLidar(_ExampleLoader):
    def __init__(self):
        self._example_data = examples.download_lidar()
        _ExampleLoader.__init__(self)


@smproxy.source(name='PyVistaTopo', label='Global Topography')
@smhint.xml('''<ShowInMenu category="%s"/>
    <RepresentationType view="RenderView" type="Surface" />''' % MENU_CAT)
class PyVistaTopo(_ExampleLoader):
    def __init__(self):
        self._example_data = examples.download_topo_global()
        _ExampleLoader.__init__(self)



"""The ``interface`` module provides functions to convert/cast between common
VTK and NumPy/Pandas data types. These methods provide a simple to use interface
for VTK data types so that users can make changes to VTK data strucutres via
Python data structures that are a bit easier to perform numerical operations
upon.
"""


__all__ = [
    'get_vtk_type',
    'convert_string_array',
    'convert_array',
    'data_frame_to_table',
    'table_to_data_frame',
    'place_array_in_table',
    'get_dtypes',
    'points_to_poly_data',
    'add_arrays_from_data_frame',
    'convert_cell_conn',
    'get_array',
    'get_data_dict',
]


__displayname__ = 'Interface'


import numpy as np
import pandas as pd
import vtk
from vtk.util import numpy_support as nps

import pyvista as pv
from pyvista.utilities import convert_string_array, get_vtk_type

from . import _helpers


def convert_array(arr, name='Data', deep=0, array_type=None, pdf=False):
    """A helper to convert a NumPy array to a vtkDataArray or vice versa

    Args:
        arr (ndarray or vtkDataArry) : A numpy array or vtkDataArry to convert
        name (str): the name of the data array for VTK
        deep (bool, int): if input is numpy array then deep copy values
        pdf (bool): if input is vtkDataArry, make a pandas DataFrame of the array

    Return:
        vtkDataArray, ndarray, or DataFrame:
            the converted array (if input is a NumPy ndaray then returns
            ``vtkDataArray`` or is input is ``vtkDataArray`` then returns NumPy
            ``ndarray``). If pdf==True and the input is ``vtkDataArry``,
            return a pandas DataFrame.

    """
    num_data = pv.convert_array(arr, name=name, deep=deep, array_type=array_type)
    if not isinstance(num_data, np.ndarray):
        return num_data
    if not pdf:
        return num_data
    return pd.DataFrame(data=num_data, columns=[arr.GetName()])



def data_frame_to_table(df, pdo=None):
    """Converts a pandas DataFrame to a vtkTable"""
    if not isinstance(df, pd.DataFrame):
        raise _helpers.PVGeoError('Input is not a pandas DataFrame')
    if pdo is None:
        pdo = pv.Table(df)
    else:
        pdo.DeepCopy(pv.Table(df))
    return pdo


def table_to_data_frame(table):
    """Converts a vtkTable to a pandas DataFrame"""
    if not isinstance(table, vtk.vtkTable):
        raise _helpers.PVGeoError('Input is not a vtkTable')
    if not isinstance(table, pv.Table):
        table = pv.Table(table)
    return table.to_pandas()


def place_array_in_table(ndarr, titles, pdo):
    """Takes a 1D/2D numpy array and makes a vtkTable of it

    Args:
        ndarr (numpy.ndarray) : The 1D/2D array to be converted to a table
        titles (list or tuple): The titles for the arrays in the table. Must
            have same number of elements as columns in input ndarray
        pdo (vtkTable) : The output data object pointer

    Return:
        vtkTable : returns the same input pdo table
    """
    # Put columns into table
    if len(np.shape(ndarr)) > 2:
        raise _helpers.PVGeoError('Input np.ndarray must be 1D or 2D to be converted to vtkTable.')
    if len(np.shape(ndarr)) == 1:
        # First check if it is an array full of tuples (varying type)
        if isinstance(ndarr[0], (tuple, np.void)):
            for i, title in enumerate(titles):
                place_array_in_table(ndarr['f%d' % i], title, pdo)
            return pv.wrap(pdo)
        # Otherwise it is just a 1D array which needs to be 2D
        else:
            ndarr = np.reshape(ndarr, (-1, 1))
    cols = np.shape(ndarr)[1]

    for i in range(cols):
        VTK_data = convert_array(ndarr[:,i])
        VTK_data.SetName(titles[i])
        pdo.AddColumn(VTK_data)
    return pv.wrap(pdo)



def get_dtypes(dtype='', endian=None):
    """This converts char dtypes and an endian to a numpy and VTK data type.

    Return:
        tuple (numpy.dtype, int):
            the numpy data type and the integer type id specified in vtkType.h
            for VTK data types
    """
    # If native `@` was chosen then do not pass an endian
    if endian == '@':
        #print('WARNING: Native endianness no longer supported for packed binary reader. Please chose `>` or `<`. This defaults to big `>`.')
        endian = ''
    # No endian specified:
    elif endian is None:
        endian = ''
    # Get numpy and VTK data types and return them both
    if dtype == 'd':
        vtktype = vtk.VTK_DOUBLE
    elif dtype == 'f':
        vtktype = vtk.VTK_FLOAT
    elif dtype == 'i':
        vtktype = vtk.VTK_INT
    else:
        raise _helpers.PVGeoError('dtype \'%s\' unknown:' % dtype)
    # Return data types
    dtype = np.dtype('%s%s' % (endian, dtype))
    return dtype, vtktype



def points_to_poly_data(points, copy_z=False):
    """Create ``vtkPolyData`` from a numpy array of XYZ points. If the points
    have more than 3 dimensions, then all dimensions after the third will be
    added as attributes. Assume the first three dimensions are the XYZ
    coordinates.

    Args:
        points (np.ndarray or pandas.DataFrame): The points and pointdata
        copy_z (bool): A flag on whether to append the z values as a PointData
            array

    Return:
        vtkPolyData : points with point-vertex cells
    """
    # This prevents an error that occurs when only one point is passed
    if points.ndim < 2:
        points = points.reshape((1,-1))
    keys = ['Field %d' % i for i in range(points.shape[1] - 3)]
    # Check if input is anything other than a NumPy array and cast it
    # e.g. you could send a Pandas dataframe
    if not isinstance(points, np.ndarray):
        if isinstance(points, pd.DataFrame):
            # If a pandas data frame, lets grab the keys
            keys = points.keys()[3::]
        points = np.array(points)
    # If points are not 3D
    if points.shape[1] < 2:
        raise RuntimeError('Points must be 3D. Try adding a third dimension of zeros.')

    atts = points[:, 3::]
    points = points[:, 0:3].astype(np.float)

    # Create polydata
    pdata = pv.PolyData(points)

    # Add attributes if given
    scalSet = False
    for i, key in enumerate(keys):
        data = convert_array(atts[:, i], name=key)
        pdata.GetPointData().AddArray(data)
        if not scalSet:
            pdata.GetPointData().SetActiveScalars(key)
            scalSet = True
    if copy_z:
        z = convert_array(points[:, 2], name='Elevation')
        pdata.GetPointData().AddArray(z)
    return pv.wrap(pdata)


def add_arrays_from_data_frame(pdo, field, df):
    """Add all of the arrays from a given data frame to an output's data"""
    for key in df.keys():
        VTK_data = convert_array(df[key].values, name=key)
        _helpers.add_array(pdo, field, VTK_data)
    return pv.wrap(pdo)



def convert_cell_conn(cell_connectivity):
    """Converts cell connectivity arrays to a cell matrix array that makes sense
    for VTK cell arrays.
    """
    cellsMat = np.concatenate(
            (
                np.ones((cell_connectivity.shape[0], 1), dtype=np.int64)*cell_connectivity.shape[1],
                cell_connectivity
            ),
            axis=1).ravel()
    return nps.numpy_to_vtk(cellsMat, deep=True, array_type=vtk.VTK_ID_TYPE)


def get_array(dataset, name, vtk_object=False):
    """Given an input dataset, this will return the named array as a NumPy array
    or a vtkDataArray if spceified
    """
    arr, field = _helpers.search_for_array(dataset, name)
    if vtk_object:
        return arr
    return convert_array(arr)


def get_data_dict(dataset, field='cell'):
    """Given an input dataset, this will return all the arrays in that object's
    cell/point/field/row data as named NumPy arrays in a dictionary.
    """
    data = {}
    for key in _helpers.get_all_array_names(dataset, field):
        data[key] = np.array(_helpers.get_numpy_array(dataset, field, key))
    return data



__all__ = [
    'PackedBinariesReader',
    'MadagascarReader',
]

__displayname__ = 'Binary/Serialized File I/O'

import warnings

import numpy as np
import vtk

from .. import _helpers, interface
from ..base import ReaderBase


class PackedBinariesReader(ReaderBase):
    """This reads in float or double data that is packed into a binary file
    format. It will treat the data as one long array and make a ``vtkTable``
    with one column of that data. The reader uses defaults to import as floats
    with native endianness. Use the Table to Uniform Grid or the Reshape Table
    filters to give more meaning to the data. We chose to use a ``vtkTable``
    object as the output of this reader because it gives us more flexibility in
    the filters we can apply to this data down the pipeline and keeps thing
    simple when using filters in this repository.
    """
    __displayname__ = 'Packed Binaries Reader'
    __category__ = 'reader'
    extensions = 'H@ bin rsf rsf@ HH npz'
    description = 'PVGeo: Packed Binaries Reader'

    def __init__(self, **kwargs):
        ReaderBase.__init__(self,
                            nOutputPorts=1, outputType='vtkTable', **kwargs)
        # Other Parameters
        self.__data_name = kwargs.get('dataname', 'Data')
        self.__dtypechar = kwargs.get('dtype', 'f')
        self.__endian = kwargs.get('endian', '')
        self.__dtype, self.__vtktype = interface.get_dtypes(dtype=self.__dtypechar, endian=self.__endian)
        # Data objects to hold the read data for access by the pipeline methods
        self.__data = []

    def _read_raw_file(self, filename):
        """Interanl helper to read the raw data from the file"""
        dtype = self.__dtype
        if dtype == np.dtype('>f'):
            # Checks if big-endian and fixes read
            dtype = np.dtype('f')
        try:
            arr = np.fromfile(filename, dtype=dtype)
        except (IOError, OSError) as fe:
            raise _helpers.PVGeoError(str(fe))
        return np.asarray(arr, dtype=self.__dtype)

    def _get_file_contents(self, idx=None):
        """Interanl helper to get all contents for all files"""
        if idx is not None:
            filenames = [self.get_file_names(idx=idx)]
        else:
            filenames = self.get_file_names()
        contents = []
        for f in filenames:
            contents.append(self._read_raw_file(f))
        if idx is not None:
            return contents[0]
        return contents


    def _read_up_front(self):
        """Should not need to be overridden
        """
        # Perform Read
        self.__data = self._get_file_contents()
        self.need_to_read(flag=False)
        return 1

    def _get_raw_data(self, idx=0):
        """This will return the proper data for the given timestep
        """
        return self.__data[idx]

    def convert_array(self, arr):
        """Converts the numpy array to a vtkDataArray
        """
        # Put raw data into vtk array
        data = interface.convert_array(arr, name=self.__data_name, deep=True, array_type=self.__vtktype)
        return data


    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to request data for current timestep
        """
        # Get output:
        output = vtk.vtkTable.GetData(outInfo)
        if self.need_to_read():
            self._read_up_front()
        # Get requested time index
        i = _helpers.get_requested_time(self, outInfo)
        # Generate the data object
        arr = self._get_raw_data(idx=i)
        data = self.convert_array(arr)
        output.AddColumn(data)
        return 1


    #### Seters and Geters ####


    def set_endian(self, endian):
        """Set the endianness of the data file.

        Args:
            endian (int or char): no preference = '' or 0, Little = 1 or `<` or Big = 2 `>`.
        """
        pos = ['', '<', '>']
        if isinstance(endian, int):
            endian = pos[endian]
        if endian != self.__endian:
            self.__endian = endian
            self.__dtype, self.__vtktype = interface.get_dtypes(dtype=self.__dtypechar, endian=self.__endian)
            self.Modified()

    def get_endian(self):
        """Get the endianness of the data file."""
        return self.__endian

    def set_data_type(self, dtype):
        """Set the data type of the binary file: `double='d'`, `float='f'`, `int='i'`
        """
        pos = ['d', 'f', 'i']
        if isinstance(dtype, int):
            dtype = pos[dtype]
        if dtype != self.__dtype:
            self.__dtypechar = dtype
            self.__dtype, self.__vtktype = interface.get_dtypes(dtype=self.__dtypechar, endian=self.__endian)
            self.Modified()

    def get_data_types(self):
        """Get the data type of the binary file"""
        return self.__dtype, self.__vtktype

    def set_data_name(self, data_name):
        """The string name of the data array generated from the inut file.
        """
        if data_name != self.__data_name:
            self.__data_name = data_name
            self.Modified(read_again=False) # Don't re-read. Just request data again


    def get_data_name(self):
        """Get name used for the data array"""
        return self.__data_name



class MadagascarReader(PackedBinariesReader):
    """This reads in float or double data that is packed into a Madagascar
    binary file format with a leader header. The reader ignores all of the ascii
    header details by searching for the sequence of three special characters:
    EOL EOL EOT and it will treat the followng binary packed data as one long
    array and make a ``vtkTable`` with one column of that data. The reader uses
    defaults to import as floats with native endianness. Use the Table to
    Uniform Grid or the Reshape Table filters to give more meaning to the data.
    We will later implement the ability to create a gridded volume from the
    header info. This reader is a quick fix for Samir. We chose to use a
    ``vtkTable`` object as the output of this reader because it gives us more
    flexibility in the filters we can apply to this data down the pipeline and
    keeps thing simple when using filters in this repository.
    `Details Here`_.

    .. _Details Here: http://www.ahay.org/wiki/RSF_Comprehensive_Description#Single-stream_RSF
    """
    __displayname__ = 'Madagascar SSRSF Reader'
    __category__ = 'reader'
    # extensions are inherrited from PackedBinariesReader
    description = 'PVGeo: Madagascar Single Stream RSF Files'

    def __init__(self, **kwargs):
        PackedBinariesReader.__init__(self, **kwargs)

    def _read_raw_file(self, filename):
        """Reads the raw data from the file for Madagascar SSRSF files"""
        dtype, vtktype = self.get_data_types()
        CTLSEQ = b'\014\014\004' # The control sequence to seperate header from data
        rpl = b''
        raw = []
        with open(filename, 'rb') as file:
            raw = file.read()
            idx = raw.find(CTLSEQ)
            if idx == -1:
                warnings.warn('This is not a single stream RSF format file. Treating entire file as packed binary data.')
            else:
                raw = raw[idx:] # deletes the header
                raw = raw.replace(CTLSEQ, rpl) # removes the control sequence
        arr = np.fromstring(raw, dtype=dtype)
        return arr



__all__ = [
    'AddCellConnToPoints',
    'PointsToTube',
    'LonLatToUTM',
    'RotatePoints',
    'ExtractPoints',
    'RotationTool',
    'ExtractCellCenters',
    'AppendCellCenters',
    'IterateOverPoints',
    'ConvertUnits',
    'BuildSurfaceFromPoints',
]

__displayname__ = 'Point/Line Sets'

from datetime import datetime

import numpy as np
import vtk
from vtk.numpy_interface import dataset_adapter as dsa

import pyvista

from .. import _helpers, interface
from ..base import FilterBase, FilterPreserveTypeBase
# improt CreateTensorMesh for its cell string parsing
from ..model_build import CreateTensorMesh

# NOTE: internal import of pyproj in LonLatToUTM


###############################################################################
#---- Cell Connectivity ----#

class AddCellConnToPoints(FilterBase):
    """This filter will add linear cell connectivity between scattered points.
    You have the option to add ``VTK_Line`` or ``VTK_PolyLine`` connectivity.
    ``VTK_Line`` connectivity makes a straight line between the points in order
    (either in the order by index or using a nearest neighbor calculation).
    The ``VTK_PolyLine`` adds a poly line connectivity between all points as
    one spline (either in the order by index or using a nearest neighbor
    calculation). Type map is specified in `vtkCellType.h`.

    **Cell Connectivity Types:**

    - 4: Poly Line
    - 3: Line
    """
    __displayname__ = 'Add Cell Connectivity to Points'
    __category__ = 'filter'

    def __init__(self, **kwargs):
        FilterBase.__init__(self,
                            nInputPorts=1, inputType='vtkPolyData',
                            nOutputPorts=1, outputType='vtkPolyData')
        # Parameters
        self.__cell_type = kwargs.get('cell_type', vtk.VTK_POLY_LINE)
        self.__usenbr = kwargs.get('nearest_nbr', False)
        self.__close_loop = kwargs.get('close_loop', False)
        self.__keep_vertices = kwargs.get('keep_vertices', False)
        self.__unique = kwargs.get('unique', False)


    def _connect_cells(self, pdi, pdo, log_time=False):
        """Internal helper to perfrom the connection"""
        # NOTE: Type map is specified in vtkCellType.h
        cell_type = self.__cell_type

        if log_time:
            start_time = datetime.now()

        # Get the Points over the NumPy interface
        pdi = pyvista.wrap(pdi)
        points = np.copy(pdi.points) # New NumPy array of poins so we dont destroy input
        if self.__unique:
            # Remove repeated points
            indexes = np.unique(points, return_index=True, axis=0)[1]
            points = np.array(points[sorted(indexes)])

        def _find_min_path(points):
            try:
                # sklearn's KDTree is faster: use it if available
                from sklearn.neighbors import KDTree as Tree
            except ImportError:
                from scipy.spatial import cKDTree as Tree
            _compute_dist = lambda pt0, pt1: np.linalg.norm(pt0-pt1)
            ind, min_dist = None, np.inf
            tree = Tree(points)
            for pt in points:
                cur_ind = tree.query([pt], k=len(points))[1].ravel()
                dist = 0.
                for i in range(len(cur_ind)-1):
                    dist += _compute_dist(points[cur_ind[i]], points[cur_ind[i+1]])
                if dist < min_dist:
                    ind = cur_ind
                    min_dist = dist
            return ind.ravel()

        if self.__usenbr:
            ind = _find_min_path(points)
        else:
            ind = np.arange(len(points), dtype=int)
        if self.__keep_vertices:
            poly = pyvista.PolyData(np.copy(points))
        else:
            poly = pyvista.PolyData()
            poly.points = np.copy(points)
        if cell_type == vtk.VTK_LINE:
            lines = np.c_[np.full(len(ind)-1, 2), ind[0:-1], ind[1:]]
            if self.__close_loop:
                app = np.append(lines, [[2, ind[-1], ind[0]],], axis=0)
                lines = app
            poly.lines = lines
        elif cell_type == vtk.VTK_POLY_LINE:
            cells = vtk.vtkCellArray()
            cell = vtk.vtkPolyLine()
            if self.__close_loop:
                cell.GetPointIds().SetNumberOfIds(len(ind) + 1)
            else:
                cell.GetPointIds().SetNumberOfIds(len(ind))
            for i in ind:
                cell.GetPointIds().SetId(i, ind[i])
            if self.__close_loop:
                cell.GetPointIds().SetId(i+1, ind[0])
            cells.InsertNextCell(cell)
            poly.SetLines(cells)
        else:
            raise _helpers.PVGeoError('Cell type ({}) not supported'.format(cell_type))
        for key, val in pdi.point_arrays.items():
            poly.point_arrays[key] = val
        pdo.DeepCopy(poly)
        if log_time:
            print("exectuted in {}".format(datetime.now() - start_time))
        return pdo

    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to generate output data object
        """
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        pdo = self.GetOutputData(outInfo, 0)
        # Perfrom task
        self._connect_cells(pdi, pdo)
        return 1


    #### Seters and Geters ####


    def set_cell_type(self, cell_type):
        """Set the cell typ by the integer id as specified in `vtkCellType.h`
        """
        if cell_type != self.__cell_type:
            self.__cell_type = cell_type
            self.Modified()

    def set_use_nearest_nbr(self, flag):
        """Set a flag on whether to a KDTree nearest neighbor
        algorithm to sort the points to before adding linear connectivity.
        """
        if flag != self.__usenbr:
            self.__usenbr = flag
            self.Modified()

    def set_use_unique_points(self, flag):
        """Set a flag on whether to only use unique points"""
        if flag != self.__unique:
            self.__unique = flag
            self.Modified()


###############################################################################


class PointsToTube(AddCellConnToPoints):
    """Takes points from a vtkPolyData object and constructs a line of those
    points then builds a polygonal tube around that line with some specified
    radius and number of sides.
    """
    __displayname__ = 'Points to Tube'
    __category__ = 'filter'

    def __init__(self, num_sides=20, radius=10.0, capping=False, **kwargs):
        AddCellConnToPoints.__init__(self, **kwargs)
        # Additional Parameters
        # NOTE: CellType should remain vtk.VTK_POLY_LINE (4) connection
        self.__numSides = num_sides
        self.__radius = radius
        self.__capping = capping


    def _connect_cells(self, pdi, pdo, log_time=False):
        """This uses the parent's ``_connect_cells()`` to build a tub around
        """
        AddCellConnToPoints._connect_cells(self, pdi, pdo, log_time=log_time)
        tube = vtk.vtkTubeFilter()
        tube.SetInputData(pdo)
        # User Defined Parameters
        tube.SetCapping(self.__capping)
        tube.SetRadius(self.__radius)
        tube.SetNumberOfSides(self.__numSides)
        # apply the filter
        tube.Update()
        pdo.ShallowCopy(tube.GetOutput())
        return pdo


    #### Seters and Geters ####


    def set_radius(self, radius):
        """Set the radius of the tube
        """
        if self.__radius != radius:
            self.__radius = radius
            self.Modified()

    def set_number_of_sides(self, num):
        """Set the number of sides (resolution) for the tube
        """
        if self.__numSides != num:
            self.__numSides = num
            self.Modified()

    def set_capping(self, flag):
        """Set a boolean flag on whether or not to cap the ends of the tube
        """
        if self.__capping != flag:
            self.__capping = flag
            self.Modified()




###############################################################################
#---- LonLat to Cartesian ----#

class LonLatToUTM(FilterPreserveTypeBase):
    """Converts Points from Lon Lat to UTM
    """
    __displayname__ = 'Lat Lon To UTM'
    __category__ = 'filter'

    def __init__(self, **kwargs):
        FilterPreserveTypeBase.__init__(self, inputType='vtkDataSet', **kwargs)
        self.__zone = 11,
        self.__ellps = 'WGS84'
        self.set_zone(kwargs.get('zone', 11)) # User defined
        self.set_ellps(kwargs.get('ellps', 'WGS84')) # User defined

    @staticmethod
    def get_available_ellps(idx=None):
        """Returns the available ellps
        """
        import pyproj
        ellps = pyproj.pj_ellps.keys()
        # Now migrate WGSXX to front so that 84 is always default
        wgs = ['WGS60','WGS66','WGS72', 'WGS84']
        for i, name in enumerate(wgs):
            oldindex = ellps.index(name)
            ellps.insert(0, ellps.pop(oldindex))
        if idx is not None:
            return ellps[idx]
        return ellps

    def __convert_2d(self, lon, lat, elev):
        """Converts 2D Lon Lat coords to 2D XY UTM points"""
        import pyproj
        p = pyproj.Proj(proj='utm', zone=self.__zone, ellps=self.__ellps)
        utm_x, utm_y = p(lon, lat)
        return np.c_[utm_x, utm_y, elev]

    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to generate output"""
        # Get input/output of Proxy
        pdi = pyvista.wrap(self.GetInputData(inInfo, 0, 0))
        pdo = self.GetOutputData(outInfo, 0)
        #### Perfrom task ####
        if not hasattr(pdi, 'points'):
            raise _helpers.PVGeoError('Input data object does not have points to convert.')
        coords = pdi.points.copy() # New NumPy array of poins so we dont destroy input
        # Now Conver the points
        points = self.__convert_2d(coords[:, 0], coords[:, 1], coords[:, 2])
        output = pdi.copy()
        output.points = points
        pdo.DeepCopy(output)
        return 1

    def set_zone(self, zone):
        """Set the UTM zone number"""
        if zone < 1 or zone > 60:
            raise _helpers.PVGeoError('Zone (%d) is invalid.' % zone)
        if self.__zone != zone:
            self.__zone = int(zone)
            self.Modified()

    def set_ellps(self, ellps):
        """Set the ellipsoid type"""
        if isinstance(ellps, int):
            ellps = self.get_available_ellps(idx=ellps)
        if not isinstance(ellps, str):
            raise _helpers.PVGeoError('Ellps must be a string.')
        if self.__ellps != ellps:
            self.__ellps = ellps
            self.Modified()


###############################################################################


class RotationTool(object):
    """A class that holds a set of methods/tools for performing and estimating
    coordinate rotations.
    """
    __displayname__ = 'Rotation Tool'
    __category__ = 'filter'

    def __init__(self, decimals=6):
        # Parameters
        self.RESOLUTION = np.pi / 3200.0
        self.DECIMALS = decimals

    @staticmethod
    def _get_rotation_matrix(theta):
        """Internal helper to generate a rotation matrix given a rotation angle"""
        xx = np.cos(theta)
        xy = -np.sin(theta)
        yx = np.sin(theta)
        yy = np.cos(theta)
        if not isinstance(theta, np.ndarray):
            return np.array([[xx, xy],
                             [yx, yy]])
        # Otherwise populate arrat manually
        mat = np.zeros((len(theta), 2, 2))
        mat[:, 0, 0] = xx
        mat[:, 0, 1] = xy
        mat[:, 1, 0] = yx
        mat[:, 1, 1] = yy
        return mat

    @staticmethod
    def rotate_around(pts, theta, origin):
        """Rotate points around an origins given an anlge on the XY plane
        """
        xarr, yarr = pts[:,0], pts[:,1]
        ox, oy = origin[0], origin[1]
        qx = ox + np.cos(theta) * (xarr - ox) - np.sin(theta) * (yarr - oy)
        qy = oy + np.sin(theta) * (xarr - ox) + np.cos(theta) * (yarr - oy)
        return np.vstack((qx, qy)).T

    @staticmethod
    def rotate(pts, theta):
        """Rotate points around (0,0,0) given an anlge on the XY plane
        """
        rot = RotationTool._get_rotation_matrix(theta)
        rotated = pts.dot(rot)
        if not isinstance(theta, np.ndarray):
            return rotated
        return np.swapaxes(rotated, 0, 1)

    @staticmethod
    def distance_between(pts):
        """Gets the distance between two points
        """
        if pts.ndim < 3:
            return np.sqrt((pts[0,0] - pts[1,0])**2 + (pts[0,1] - pts[1,1])**2)
        return np.sqrt((pts[:, 0,0] - pts[:, 1,0])**2 + (pts[:, 0,1] - pts[:, 1,1])**2)

    @staticmethod
    def cos_between(pts):
        """Gets the cosine between two points
        """
        if pts.ndim < 3:
            xdiff = abs(pts[0,0] - pts[1,0])
            dist = RotationTool.distance_between(pts)
            return np.arccos(xdiff/dist)
        # Otherwise we have a set of points
        xdiff = abs(pts[:, 0,0] - pts[:, 1,0])
        dist = RotationTool.distance_between(pts)
        return np.arccos(xdiff/dist)

    @staticmethod
    def sin_between(pts):
        """Calculate the sin angle between two points"""
        ydiff = abs(pts[0,1] - pts[1,1])
        dist = RotationTool.distance_between(pts)
        return np.arcsin(ydiff/dist)

    @staticmethod
    def rotation_matrix(vector_orig, vector_fin):
        """Calculate the rotation matrix required to rotate from one vector to another.
        For the rotation of one vector to another, there are an infinit series of rotation matrices
        possible.  Due to axially symmetry, the rotation axis can be any vector lying in the symmetry
        plane between the two vectors.  Hence the axis-angle convention will be used to construct the
        matrix with the rotation axis defined as the cross product of the two vectors.  The rotation
        angle is the arccosine of the dot product of the two unit vectors.
        Given a unit vector parallel to the rotation axis, w = [x, y, z] and the rotation angle a,
        the rotation matrix R is::
                |  1 + (1-cos(a))*(x*x-1)   -z*sin(a)+(1-cos(a))*x*y   y*sin(a)+(1-cos(a))*x*z |
            R = |  z*sin(a)+(1-cos(a))*x*y   1 + (1-cos(a))*(y*y-1)   -x*sin(a)+(1-cos(a))*y*z |
                | -y*sin(a)+(1-cos(a))*x*z   x*sin(a)+(1-cos(a))*y*z   1 + (1-cos(a))*(z*z-1)  |

        Args:
            vector_orig (umpy array, len 3): The unrotated vector defined in the reference frame.
            vector_fin (numpy array, len 3): The rotated vector defined in the reference frame.

        Note:
            This code was adopted from `printipi`_ under the MIT license.

        .. _printipi: https://github.com/Wallacoloo/printipi/blob/master/util/rotation_matrix.py
        """
        from math import acos, cos, sin
        from numpy import cross, dot
        from numpy.linalg import norm

        R = np.zeros((3,3))

        # Convert the vectors to unit vectors.
        vector_orig = vector_orig / norm(vector_orig)
        vector_fin = vector_fin / norm(vector_fin)

        # The rotation axis (normalised).
        axis = cross(vector_orig, vector_fin)
        axis_len = norm(axis)
        if axis_len != 0.0:
            axis = axis / axis_len

        # Alias the axis coordinates.
        x = axis[0]
        y = axis[1]
        z = axis[2]

        # The rotation angle.
        angle = acos(dot(vector_orig, vector_fin))

        # Trig functions (only need to do this maths once!).
        ca = cos(angle)
        sa = sin(angle)

        # Calculate the rotation matrix elements.
        R[0,0] = 1.0 + (1.0 - ca)*(x**2 - 1.0)
        R[0,1] = -z*sa + (1.0 - ca)*x*y
        R[0,2] = y*sa + (1.0 - ca)*x*z
        R[1,0] = z*sa+(1.0 - ca)*x*y
        R[1,1] = 1.0 + (1.0 - ca)*(y**2 - 1.0)
        R[1,2] = -x*sa+(1.0 - ca)*y*z
        R[2,0] = -y*sa+(1.0 - ca)*x*z
        R[2,1] = x*sa+(1.0 - ca)*y*z
        R[2,2] = 1.0 + (1.0 - ca)*(z**2 - 1.0)
        return R


    # def _converge_angle2(self, pt1, pt2):
    #     """internal use only: pts should only be a two neighboring points"""
    #     # Make the theta range up to 90 degrees to rotate points through
    #     #- angles = [0.0, 90.0)
    #     angles = np.arange(0.0, np.pi/2, self.RESOLUTION)
    #     pts = self.rotate(np.vstack((pt1, pt2)), angles)
    #     # Get the angles between the points
    #     c = self.cos_between(pts)
    #     dist = self.distance_between(pts)
    #
    #     # Find angles that satisfy grid conditions
    #     xidx = np.argwhere(abs(c - np.pi/2.0) < (1 * 10**-self.DECIMALS))
    #     yidx = np.argwhere(abs(c - 0.0) < (1 * 10**-self.DECIMALS))
    #     if len(xidx) == 1 and len(yidx) == 0:
    #         return 0, np.pi/2-angles[xidx], dist[xidx]
    #         #return angles[xidx[0][0]]
    #     elif len(yidx) == 1 and len(xidx) == 0:
    #         return 1, np.pi/2-angles[yidx], dist[yidx]
    #     else:
    #         raise _helpers.PVGeoError('No angle found')


    def _converge_angle(self, pt1, pt2):
        """Internal use only: pts should only be a two neighboring points.
        """
        # Make the theta range up to 90 degrees to rotate points through
        #- angles = [0.0, 90.0)
        angles = np.arange(0.0, np.pi/2, self.RESOLUTION)
        # nang = len(angles) # Number of rotations

        # if pt1.ndim == pt2.ndim == 3:
        #     # uh-oh
        #     raise RuntimeError()
        pts = np.vstack((pt1, pt2))

        rotated = self.rotate(pts, angles) # Points rotated for all angles
        cosbtw = self.cos_between(rotated)
        distbtw = self.distance_between(rotated)
        # Now find minimum

        # X axis
        xmin = np.argwhere(np.abs(cosbtw - np.pi/2.0) < (1 * 10**-self.DECIMALS)).flatten()
        ymin = np.argwhere(np.abs(cosbtw - 0.0) < (1 * 10**-self.DECIMALS)).flatten()

        # Protection to make sure we can converge
        if len(xmin) == 0 and len(ymin) == 0:
            # Uh-oh... lets decrease the precision
            #- lets try again with lower precision
            self.DECIMALS -= 1
            if self.DECIMALS < 0:
                self.DECIMALS = 0
                raise _helpers.PVGeoError('No angle found.')
            return self._converge_angle(pt1, pt2)

        # Figure out of the two points share the x axis or y axis and return
        if len(xmin) > 0 and len(ymin) > 0:
            raise RuntimeError('Invalid solution')
        elif len(xmin) > 0:
            xidx = np.mean(xmin, dtype=int)
            return 0, angles[xidx], distbtw[xidx]
        elif len(ymin) > 0:
            yidx = np.mean(ymin, dtype=int)
            return 1, angles[yidx], distbtw[yidx]
        # No solution found.
        raise _helpers.PVGeoError('No angle found. Precision too low/high.')


    def _estimate_angle_and_spacing(self, pts, sample=0.5):
        """internal use only
        """
        try:
            # sklearn's KDTree is faster: use it if available
            from sklearn.neighbors import KDTree as Tree
        except ImportError:
            from scipy.spatial import cKDTree as Tree
        # Creat the indexing range for searching the points:
        num = len(pts)
        rng = np.linspace(0, num-1, num=num, dtype=int)
        N = int(num*sample) + 1
        rng = np.random.choice(rng, N)
        angles = np.zeros(len(rng))
        tree = Tree(pts)
        distances = [[],[]]

        #######################################################################
        #######################################################################
        # Find nearest point
        distall, ptsiall = tree.query(pts, k=2)
        pt1all, pt2all = pts[ptsiall[:, 0]], pts[ptsiall[:, 1]]
        #######################################################################
        idx = 0
        for i in rng:
            # OPTIMIZE
            ax, angles[idx], dist = self._converge_angle(pt1all[i], pt2all[i])
            distances[ax].append(dist)
            idx += 1
        #######################################################################
        #TODO??? angles, distances = self._converge_angle(pt1all, pt2all)
        #######################################################################
        #######################################################################
        dx, dy = distances[0], distances[1]
        if len(dx) == 0:
            dx = dy
        elif len(dy) == 0:
            dy = dx
        TOLERANCE = np.min(np.append(dx, dy)) / 2.0
        angle = np.average(np.unique(angles))
        dx = np.unique(np.around(dx / TOLERANCE)) * TOLERANCE
        dy = np.unique(np.around(dy / TOLERANCE)) * TOLERANCE

        # Now round to decimals
        dx = np.around(dx, self.DECIMALS)
        dy = np.around(dx, self.DECIMALS)

        # print('Recovered: ', dx, dy)
        return angle, dx[0], dy[0]


    def estimate_and_rotate(self, x, y, z):
        """A method to estimate the rotation of a set of points and correct
        that rotation on the XY plane
        """
        if not (len(x) == len(y) == len(z)):
            raise AssertionError('Must have same number of coordinates for all components.')
        idxs = np.argwhere(z == z[0])
        pts = np.hstack((x[idxs], y[idxs]))
        angle, dx, dy = self._estimate_angle_and_spacing(pts)
        inv = self.rotate(np.vstack((x, y)).T, angle)
        return inv[:,0], inv[:,1], z, dx, dy, angle



#---- Coordinate Rotations ----#

class RotatePoints(FilterBase):
    """Rotates XYZ coordinates in `vtkPolyData` around an origin at a given
    angle on the XY plane.
    """
    __displayname__ = 'Rotate Points'
    __category__ = 'filter'

    def __init__(self, angle=45.0, origin=None, use_corner=True):
        FilterBase.__init__(self,
                            nInputPorts=1, inputType='vtkPolyData',
                            nOutputPorts=1, outputType='vtkPolyData')
        # Parameters
        self.__angle = angle
        if origin is None:
            origin = [0.0, 0.0]
        self.__origin = origin
        self.__use_corner = use_corner

    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to generate output.
        """
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        pdo = self.GetOutputData(outInfo, 0)
        #### Perfrom task ####
        # Get the Points over the NumPy interface
        wpdi = dsa.WrapDataObject(pdi) # NumPy wrapped input
        points = np.array(wpdi.Points) # New NumPy array of poins so we dont destroy input
        origin = self.__origin
        if self.__use_corner:
            idx = np.argmin(points[:,0])
            origin = [points[idx,0], points[idx,1]]
        points[:,0:2] = RotationTool.rotate_around(points[:,0:2], self.__angle, origin)
        pdo.DeepCopy(pdi)
        pts = pdo.GetPoints()
        for i, pt in enumerate(points):
            pts.SetPoint(i, pt)
        return 1

    def set_rotation_degrees(self, theta):
        """Sets the rotational angle in degrees.
        """
        theta = np.deg2rad(theta)
        if self.__angle != theta:
            self.__angle = theta
            self.Modified()

    def set_origin(self, xo, yo):
        """Sets the origin to perform the rotate around.
        """
        if self.__origin != [xo, yo]:
            self.__origin = [xo, yo]
            self.Modified()

    def set_use_corner(self, flag):
        """A flag to use a corner of the input data set as the rotational
        origin.
        """
        if self.__use_corner != flag:
            self.__use_corner = flag
            self.Modified()



###############################################################################

class ExtractPoints(FilterBase):
    """Extracts XYZ coordinates and point/cell data from an input ``vtkDataSet``
    """
    __displayname__ = 'Extract Points'
    __category__ = 'filter'

    def __init__(self):
        FilterBase.__init__(self,
                            nInputPorts=1, inputType='vtkDataSet',
                            nOutputPorts=1, outputType='vtkPolyData')

    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to generate output"""
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        pdo = self.GetOutputData(outInfo, 0)
        #### Perfrom task ####
        # Get the Points over the NumPy interface
        wpdi = dsa.WrapDataObject(pdi) # NumPy wrapped input
        if not hasattr(wpdi, 'Points'):
            raise _helpers.PVGeoError('Input data object does not have XYZ points.')
        points = np.array(wpdi.Points) # New NumPy array of poins so we dont destroy input
        # Now transfer data
        f = vtk.vtkCellDataToPointData()
        f.SetInputData(pdi)
        f.Update()
        d = f.GetOutput()
        pdo.ShallowCopy(interface.points_to_poly_data(points))
        _helpers.copy_arrays_to_point_data(d, pdo, 0) # 0 is point data
        return 1



class ExtractCellCenters(FilterBase):
    __displayname__ = 'Extract Cell Centers'
    __category__ = 'filter'

    def __init__(self, **kwargs):
        FilterBase.__init__(self, nInputPorts=1, inputType='vtkDataSet',
                            nOutputPorts=1, outputType='vtkPolyData', **kwargs)

    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to generate output"""
        pdi = self.GetInputData(inInfo, 0, 0)
        pdo = self.GetOutputData(outInfo, 0)
        # Find cell centers
        filt = vtk.vtkCellCenters()
        filt.SetInputDataObject(pdi)
        filt.Update()

        centers = dsa.WrapDataObject(filt.GetOutput()).Points
        # Get CellData
        wpdi = dsa.WrapDataObject(pdi)
        celldata = wpdi.CellData
        keys = celldata.keys()

        # Make poly data of Cell centers:
        pdo.DeepCopy(interface.points_to_poly_data(centers))
        for i, name in enumerate(keys):
            pdo.GetPointData().AddArray(pdi.GetCellData().GetArray(name))
        return 1


class AppendCellCenters(FilterPreserveTypeBase):
    __displayname__ = 'Append Cell Centers'
    __category__ = 'filter'

    def __init__(self, **kwargs):
        FilterPreserveTypeBase.__init__(self, **kwargs)

    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to generate output"""
        pdi = self.GetInputData(inInfo, 0, 0)
        pdo = self.GetOutputData(outInfo, 0)
        # Find cell centers
        filt = vtk.vtkCellCenters()
        filt.SetInputDataObject(pdi)
        filt.Update()

        # I use the dataset adapter/numpy interface because its easy
        centers = dsa.WrapDataObject(filt.GetOutput()).Points
        centers = interface.convert_array(centers)
        centers.SetName('Cell Centers')

        # Copy input data and add cell centers as tuple array
        pdo.DeepCopy(pdi)
        pdo.GetCellData().AddArray(centers)

        return 1




class IterateOverPoints(FilterBase):
    """Iterate over points in a time varying manner.
    """
    __displayname__ = 'Iterate Over Points'
    __category__ = 'filter'

    def __init__(self, dt=1.0):
        FilterBase.__init__(self, nInputPorts=1, inputType='vtkPolyData',
                            nOutputPorts=1, outputType='vtkPolyData')
        # Parameters
        self.__dt = dt
        self.__timesteps = None
        self.__original = 2
        self.__tindex = None
        self.__n = 2
        self.__decimate = 100
        # The point/normal that gets updated on every iteration
        self.__point = (0.0, 0.0, 0.0)
        self.__normal = (1.0, 0.0, 0.0)


    def _update_time_steps(self):
        """For internal use only
        """
        self.__timesteps = _helpers.update_time_steps(self, self.__n, self.__dt)


    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to generate output"""
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        # Get number of points
        pdo = self.GetOutputData(outInfo, 0)
        #### Perfrom task ####
        # Get the Points over the NumPy interface
        # wpdi = dsa.WrapDataObject(pdi) # NumPy wrapped input
        # Get requested time index
        i = _helpers.get_requested_time(self, outInfo)
        # Now grab point at this timestep
        pt = pdi.GetPoints().GetPoint(self.__tindex[i])
        # Calculate normal
        pts1 = self.__point
        pts2 = pt
        x1, y1, z1 = pts1[0], pts1[1], pts1[2]
        x2, y2, z2 = pts2[0], pts2[1], pts2[2]
        normal = [x2-x1, y2-y1, z2-z1]
        self.__point = pt
        self.__normal = normal
        poly = interface.points_to_poly_data(np.array(pt))
        pdo.ShallowCopy(poly)
        return 1


    def RequestInformation(self, request, inInfo, outInfo):
        """Used by pipeline to set the time information
        """
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        # Get number of points
        self.__original = pdi.GetNumberOfPoints()
        self.set_decimate(self.__decimate)
        # register time:
        self._update_time_steps()
        return 1

    #### Public Getters / Setters ####

    def set_decimate(self, percent):
        """Set the percent (1 to 100) to decimate
        """
        if percent > 100 or percent < 1:
            return
        self.__decimate = percent
        self.__n = int(self.__original * (percent/100.0))
        self.__tindex = np.linspace(0, self.__original-1, self.__n, dtype=int)
        self._update_time_steps()
        self.Modified()

    def set_time_delta(self, dt):
        """
        Set the time step interval in seconds
        """
        if self.__dt != dt:
            self.__dt = dt
            self._update_time_steps()
            self.Modified()

    def get_time_step_values(self):
        """Use this in ParaView decorator to register timesteps
        """
        return self.__timesteps.tolist() if self.__timesteps is not None else None

    def get_point(self):
        """Get the current point"""
        return list(self.__point)

    def get_normal(self):
        """Get the current normal vector"""
        return list(self.__normal)




class ConvertUnits(FilterPreserveTypeBase):
    """Convert points in an input data object to from meters to feet or vice versa.
    This simply uses a ``vtkTransformFilter`` and scales input data object with
    common conversions.
    """
    __displayname__ = 'Convert XYZ Units'
    __category__ = 'filter'

    def __init__(self, conversion='meter_to_feet', **kwargs):
        FilterPreserveTypeBase.__init__(self, **kwargs)
        self.__conversion = conversion

    @staticmethod
    def lookup_conversions(get_keys=False):
        """All Available conversions

        Return:
            dict: dictionary of conversion units
        """
        convs = dict(
            meter_to_feet=3.2808399,
            feet_to_meter=1/3.2808399,
        )
        if get_keys:
            return convs.keys()
        return convs

    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to generate output"""
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        # Get number of points
        pdo = self.GetOutputData(outInfo, 0)
        #### Perfrom task ####
        filt = vtk.vtkTransformFilter()
        trans = vtk.vtkTransform()
        trans.Scale(self.get_conversion(), self.get_conversion(), self.get_conversion())
        filt.SetTransform(trans)
        filt.SetInputDataObject(pdi)
        filt.Update()
        scaled = filt.GetOutputDataObject(0)
        pdo.DeepCopy(scaled)
        return 1



    def set_conversion(self, key):
        """Set the conversion via a lookup table"""
        convs = self.lookup_conversions()
        if isinstance(key, str):
            if key.lower() not in convs.keys():
                raise _helpers.PVGeoError('Converion `%s` not available.' % key)
        elif isinstance(key, int):
            key = convs.keys()[key]
        if self.__conversion != key:
            self.__conversion = key
            self.Modified()
        return 1

    def get_conversion(self):
        """Get the conversion value"""
        convs = self.lookup_conversions()
        return convs[self.__conversion]





class BuildSurfaceFromPoints(FilterBase):
    """From the sorted x, y, and z station locations in the input PolyData,
    create a surface to project down from the line of those points. Use the
    Z cells to control the size of the mesh surface
    """
    __displayname__ = 'Build Surface From Points'
    __category__ = 'filter'

    def __init__(self, **kwargs):
        FilterBase.__init__(self, inputType='vtkPolyData',
                            outputType='vtkUnstructuredGrid', **kwargs)
        self.__zcoords = CreateTensorMesh._read_cell_line('0. 50.')
        zcoords = kwargs.get('zcoords', self.__zcoords)
        if not isinstance(zcoords, (str, list, tuple, np.ndarray)):
            raise TypeError('zcoords of bad type.')
        if isinstance(zcoords, str):
            self.set_z_coords_str(zcoords)
        else:
            self.set_z_coords(zcoords)


    @staticmethod
    def create_surface(points, z_range):
        """From the sorted x, y, and z station locations, create a surface
        to display a seismic recording/migration on in space. The result is
        defined in the X,Y,Z-z_range 3D space.

        The z_range should be treated as relative coordinates to the values
        given on the third column of the points array. If you want the values
        in the z_range to be treated as the absolute coordinates, simply
        do not pass any Z values in the points array - if points is N by 2,
        then the values in z_range will be inferred as absolute.

        Args:
            points (np.ndarray): array-like of the station x and y locations
            (npts by 2-3) z_range (np.ndarray): The linear space of the z
            dimension. This will be filled out for every station location.

        Return:
            pyvista.UnstructuredGrid
        """
        if hasattr(points, 'values'):
            # This will extract data from pandas dataframes if those are given
            points = points.values
        points = np.array(points)
        z_range = np.array(z_range)
        xloc = points[:,0]
        yloc = points[:,1]
        if points.shape[1] > 2:
            zloc = points[:,2]
        else:
            val = np.nanmax(z_range)
            z_range = val - np.flip(z_range)
            zloc = np.full(xloc.shape, val)
        if not len(xloc) == len(yloc) == len(zloc):
            raise AssertionError('Coordinate shapes do not match.')
        nt = len(xloc)
        ns = len(z_range)

        # Extrapolate points to a 2D surface
        # repeat the XY locations across
        points = np.repeat(np.c_[xloc,yloc,zloc], ns, axis=0)
        # repeat the Z locations across
        tp = np.repeat(z_range.reshape((-1, len(z_range))), nt, axis=0)
        tp = zloc[:,None] - tp
        points[:,-1] = tp.ravel()

        # Create cell indices for that surface
        indexes = np.array(range(0, (nt*ns)))
        indexes = np.reshape(indexes, (nt, ns))

        # Define the cell connectivity on the surface
        cellConn = np.zeros((nt-1, ns-1, 4), dtype=np.int)
        cellConn[:,:,0] = indexes[:-1, :-1]
        cellConn[:,:,1] = indexes[1:, :-1]
        cellConn[:,:,2] = indexes[1:, 1:]
        cellConn[:,:,3] = indexes[:-1, 1:]
        cellConn = cellConn.reshape((ns-1)*(nt-1), 4)
        cells = vtk.vtkCellArray()
        cells.SetNumberOfCells(cellConn.shape[0])
        cells.SetCells(cellConn.shape[0], interface.convert_cell_conn(cellConn))

        # Produce the output
        output = pyvista.UnstructuredGrid()
        output.points = points
        output.SetCells(vtk.VTK_QUAD, cells)
        return output


    def RequestData(self, request, inInfo, outInfo):
        """Execute on pipeline"""
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        # Get number of points
        pdo = self.GetOutputData(outInfo, 0)
        #### Perfrom task ####
        data = pyvista.wrap(pdi)
        output = BuildSurfaceFromPoints.create_surface(data.points, np.array(self.__zcoords))
        pdo.DeepCopy(output)
        return 1

    def set_z_coords(self, zcoords):
        """Set the spacings for the cells in the Z direction

        Args:
            zcoords (list or np.array(floats)): the spacings along the Z-axis"""
        if len(zcoords) != len(self.__zcoords) or not np.allclose(self.__zcoords, zcoords):
            self.__zcoords = zcoords
            self.Modified()

    def set_z_coords_str(self, zcoordstr):
        """Set the spacings for the cells in the Z direction

        Args:
            zcoordstr (str)  : the spacings along the Z-axis in the UBC style"""
        zcoords = CreateTensorMesh._read_cell_line(zcoordstr)
        self.set_z_coords(zcoords)



__all__ = [
    'ArrayMath',
    'NormalizeArray',
    'PercentThreshold',
    'ArraysToRGBA',
]

__displayname__ = 'Math Operations'

import numpy as np
import vtk
from vtk.numpy_interface import dataset_adapter as dsa

import pyvista as pv

from .. import _helpers, interface
from ..base import FilterBase, FilterPreserveTypeBase

###############################################################################

#---- ArrayMath ----#
class ArrayMath(FilterPreserveTypeBase):
    """This filter allows the user to select two input data arrays on which to
    perfrom math operations. The input arrays are used in their order of
    selection for the operations.

    Args:
        multiplier (float) : a static shifter/scale factor across the array
            after normalization.

        new_name (str): The new array's string name

        operation (str, int, or callable): The operation as a string key, int
            index, or callable method

    **Available Math Operations:**

    - `add`: This adds the two data arrays together
    - `subtract`: This subtracts input array 2 from input array 1
    - `multiply`: Multiplies the two data arrays together
    - `divide`: Divide input array 1 by input array 2 (arr1/arr2)
    - `correlate`: Use `np.correlate(arr1, arr2, mode='same')`
    """
    __displayname__ = 'Array Math'
    __category__ = 'filter'

    def __init__(self, **kwargs):
        FilterPreserveTypeBase.__init__(self)
        # Parameters:
        self.__multiplier = kwargs.get('multiplier', 1.0)
        self.__new_name = kwargs.get('new_name', 'Mathed Up')
        self.__input_array_1 = [None, None]
        self.__input_array_2 = [None, None]
        # Convert operation to callable method
        op = kwargs.get('operation', 'add')
        if isinstance(op, (str, int)):
            op = self.get_operation(op)
        self.__operation = op


    @staticmethod
    def _correlate(arr1, arr2):
        """Use ``np.correlate()`` on ``mode='same'`` on two selected arrays
        from one input.
        """
        return np.correlate(arr1, arr2, mode='same')

    @staticmethod
    def _multiply(arr1, arr2):
        """Mutlipies two input NumPy arrays"""
        return arr1*arr2

    @staticmethod
    def _divide(arr1, arr2):
        """Divides two input NumPy arrays"""
        return arr1/arr2

    @staticmethod
    def _add(arr1, arr2):
        """Adds two input NumPy arrays"""
        return arr1+arr2

    @staticmethod
    def _subtract(arr1, arr2):
        """Subtracts two input NumPy arrays"""
        return arr1-arr2

    @staticmethod
    def get_operations():
        """Returns the math operation methods as callable objects in a
        dictionary
        """
        ops = dict(
            add=ArrayMath._add,
            subtract=ArrayMath._subtract,
            multiply=ArrayMath._multiply,
            divide=ArrayMath._divide,
            correlate=ArrayMath._correlate,
        )
        return ops

    @staticmethod
    def get_operation_names():
        """Gets a list of the math operation keys

        Return:
            list(str): the keys for getting the math operations
        """
        ops = ArrayMath.get_operations()
        return list(ops.keys())

    @staticmethod
    def get_operation(idx):
        """Gets a math operation based on an index in the keys

        Return:
            callable: the math operation method
        """
        if isinstance(idx, str):
            return ArrayMath.get_operations()[idx]
        n = ArrayMath.get_operation_names()[idx]
        return ArrayMath.get_operations()[n]


    def _math_up(self, pdi, pdo):
        """Make sure to pass array names and integer associated fields.
        Use helpers to get these properties.
        """
        if pdo is None:
            # TODO: test this
            pdo = pdi.DeepCopy()
        # Get the input arrays
        field1, name1 = self.__input_array_1[0], self.__input_array_1[1]
        field2, name2 = self.__input_array_2[0], self.__input_array_2[1]
        wpdi = dsa.WrapDataObject(pdi)
        arr1 = _helpers.get_numpy_array(wpdi, field1, name1)
        arr2 = _helpers.get_numpy_array(wpdi, field2, name2)
        # Perform Math Operation
        carr = self.__operation(arr1, arr2)
        # apply the multiplier
        carr *= self.__multiplier
        # If no name given for data by user, use operator name
        new_name = self.__new_name
        if new_name == '':
            new_name = 'Mathed Up'
        # Convert to a VTK array
        c = interface.convert_array(carr, name=new_name)
        # Build output
        pdo.DeepCopy(pdi)
        pdo = _helpers.add_array(pdo, field1, c)
        return pdo


    #### Algorithm Methods ####


    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to perfrom operation and generate output
        """
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        pdo = self.GetOutputData(outInfo, 0)
        # Perfrom task
        self._math_up(pdi, pdo)
        return 1


    #### Seters and Geters ####


    def _set_input_array_1(self, field, name):
        """Set 1st input array by name and field"""
        if self.__input_array_1[0] != field:
            self.__input_array_1[0] = field
            self.Modified()
        if self.__input_array_1[1] != name:
            self.__input_array_1[1] = name
            self.Modified()

    def _set_input_array_2(self, field, name):
        """Set 2nd input array by name and field"""
        if self.__input_array_2[0] != field:
            self.__input_array_2[0] = field
            self.Modified()
        if self.__input_array_2[1] != name:
            self.__input_array_2[1] = name
            self.Modified()

    def SetInputArrayToProcess(self, idx, port, connection, field, name):
        """Used to set the input array(s)

        Args:
            idx (int): the index of the array to process
            port (int): input port (use 0 if unsure)
            connection (int): the connection on the port (use 0 if unsure)
            field (int): the array field (0 for points, 1 for cells, 2 for
                field, and 6 for row)
            name (int): the name of the array
        """
        if idx == 0:
            self._set_input_array_1(field, name)
        elif idx == 1:
            self._set_input_array_2(field, name)
        else:
            raise _helpers.PVGeoError('SetInputArrayToProcess() do not know how to handle idx: %d' % idx)
        return 1

    def apply(self, input_data_object, array_name_0, array_name_1):
        """Run the algorith on an input data object, specifying array names"""
        self.SetInputDataObject(input_data_object)
        arr0, field0 = _helpers.search_for_array(input_data_object, array_name_0)
        arr1, field1 = _helpers.search_for_array(input_data_object, array_name_1)
        self.SetInputArrayToProcess(0, 0, 0, field0, array_name_0)
        self.SetInputArrayToProcess(1, 0, 0, field1, array_name_1)
        self.Update()
        return pv.wrap(self.GetOutput())

    def set_multiplier(self, val):
        """This is a static shifter/scale factor across the array after
        normalization.
        """
        if self.__multiplier != val:
            self.__multiplier = val
            self.Modified()

    def get_multiplier(self):
        """Return the set multiplier/scalar
        """
        return self.__multiplier

    def set_new_array_name(self, name):
        """Give the new array a meaningful name.
        """
        if self.__new_name != name:
            self.__new_name = name
            self.Modified()

    def get_new_array_name(self):
        """Get the name used for the new array"""
        return self.__new_name

    def set_operation(self, op):
        """Set the math operation to perform

        Args:
            op (str, int, or callable): The operation as a string key, int
            index, or callable method

        Note:
            This can accept a callable method to set a custom operation as long
            as its signature is: ``<callable>(arr1, arr2)``
        """
        if isinstance(op, str):
            op = ArrayMath.get_operations()[op]
        elif isinstance(op, int):
            op = ArrayMath.get_operation(op)
        if self.__operation != op:
            self.__operation = op
            self.Modified()



###############################################################################

#---- Normalizations ----#

class NormalizeArray(FilterPreserveTypeBase):
    """This filter allows the user to select an array from the input data set
    to be normalized. The filter will append another array to that data set for
    the output. The user can specify how they want to rename the array, can
    choose a multiplier, and can choose from several types of common
    normalizations (more functionality added as requested).

    Args:
        multiplier (float) : a static shifter/scale factor across the array
            after normalization.

        new_name (str): The new array's string name

        absolute (bool):

        normalization (str, int, or callable): The operation as a string key,
            integer index, or callable method

    **Normalization Types:**

    - `feature_scale`: Feature Scale
    - `standard_score`: tandard Score
    - `log10`: Natural Log
    - `natural_log`: Log Base 10
    - `just_multiply`: Only Multiply by Multiplier
    """
    __displayname__ = 'Normalize Array'
    __category__ = 'filter'

    def __init__(self, **kwargs):
        FilterPreserveTypeBase.__init__(self)
        # Parameters:
        self.__multiplier = kwargs.get('multiplier', 1.0)
        self.__new_name = kwargs.get('new_name', 'Normalized')
        self.__absolute = kwargs.get('absolute', False)
        self.__input_array = [None, None]
        # Convert operation to callable method
        op = kwargs.get('normalization', 'feature_scale')
        if isinstance(op, (str, int)):
            op = self.get_normalization(op)
        self.__normalization = op
        self.__shift = 0.0


    #### Array normalization methods ####


    @staticmethod
    def _pass_array(arr):
        """Cast an input array as a NumPy array"""
        return np.array(arr)

    @staticmethod
    def _feature_scale(arr, rng=None):
        """Returns feature scale normalization of input array"""
        # TODO: implement ability to use custom range
        if rng is not None:
            mi = rng[0]
            ma = rng[1]
        else:
            mi = np.nanmin(arr)
            ma = np.nanmax(arr)
        return (arr - mi) / (ma - mi)

    @staticmethod
    def _standard_score(arr):
        """Returns tandard score normalization of input array"""
        return (arr - np.mean(arr)) / (np.std(arr))

    @staticmethod
    def _log10(arr):
        """Returns log base 10 of input array"""
        return np.log10(arr)

    @staticmethod
    def _log_nat(arr):
        """Returns natural logarithm of input array"""
        return np.log(arr)

    @staticmethod
    def get_normalizations():
        """All Available normalizations

        Return:
            dict: dictionary of callable methods for normalizing an array
        """
        ops = dict(
            feature_scale=NormalizeArray._feature_scale,
            standard_score=NormalizeArray._standard_score,
            log10=NormalizeArray._log10,
            natural_log=NormalizeArray._log_nat,
            just_multiply=NormalizeArray._pass_array,
        )
        return ops

    @staticmethod
    def get_normalization_names():
        """Gets a list of the normalization keys

        Return:
            list(str): the keys for getting the normalizations
        """
        ops = NormalizeArray.get_normalizations()
        return list(ops.keys())

    @staticmethod
    def get_normalization(idx):
        """Gets a normalization based on an index in the keys

        Return:
            callable: the normalization method
        """
        if isinstance(idx, str):
            return NormalizeArray.get_normalizations()[idx]
        n = NormalizeArray.get_normalization_names()[idx]
        return NormalizeArray.get_normalizations()[n]

    @staticmethod
    def get_array_range(pdi, field, name):
        """Returns a tuple of the range for a ``vtkDataArray`` on a
        ``vtkDataObject``.
        """
        wpdi = dsa.WrapDataObject(pdi)
        arr = _helpers.get_numpy_array(wpdi, field, name)
        arr = np.array(arr)
        return (np.nanmin(arr), np.nanmax(arr))


    def _normalize(self, pdi, pdo):
        """Perform normalize on a data array for any given VTK data object.
        """
        # Get input array
        field, name = self.__input_array[0], self.__input_array[1]
        #self.__range = NormalizeArray.get_array_range(pdi, field, name)
        wpdi = dsa.WrapDataObject(pdi)
        arr = _helpers.get_numpy_array(wpdi, field, name)
        arr = np.array(arr, dtype=float)
        # Take absolute value?
        if self.__absolute:
            arr = np.abs(arr)
        arr += self.__shift
        # Perform normalization scheme
        arr = self.__normalization(arr)
        # apply the multiplier
        arr *= self.__multiplier
        # If no name given for data by user, use operator name
        new_name = self.__new_name
        if new_name == '':
            new_name = 'Normalized ' + name
        # Convert to VTK array
        c = interface.convert_array(arr, name=new_name)
        # Build output
        pdo.DeepCopy(pdi)
        pdo = _helpers.add_array(pdo, field, c)
        return pdo

    #### Algorithm Methods ####


    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to generate output
        """
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        pdo = self.GetOutputData(outInfo, 0)
        # Perfrom task
        self._normalize(pdi, pdo)
        return 1

    #### Seters and Geters ####


    def SetInputArrayToProcess(self, idx, port, connection, field, name):
        """Used to set the input array(s)

        Args:
            idx (int): the index of the array to process
            port (int): input port (use 0 if unsure)
            connection (int): the connection on the port (use 0 if unsure)
            field (int): the array field (0 for points, 1 for cells, 2 for
                field, and 6 for row)
            name (int): the name of the array
        """
        if self.__input_array[0] != field:
            self.__input_array[0] = field
            self.Modified()
        if self.__input_array[1] != name:
            self.__input_array[1] = name
            self.Modified()
        return 1

    def apply(self, input_data_object, array_name):
        """Run the algorithm on an input data object, specifying the array"""
        self.SetInputDataObject(input_data_object)
        arr, field = _helpers.search_for_array(input_data_object, array_name)
        self.SetInputArrayToProcess(0, 0, 0, field, array_name)
        self.Update()
        return pv.wrap(self.GetOutput())

    def set_multiplier(self, val):
        """This is a static shifter/scale factor across the array after
        normalization.
        """
        if self.__multiplier != val:
            self.__multiplier = val
            self.Modified()

    def get_multiplier(self):
        """Return the set multiplier/scalar
        """
        return self.__multiplier


    def set_new_array_name(self, name):
        """Give the new array a meaningful name.
        """
        if self.__new_name != name:
            self.__new_name = name
            self.Modified()

    def get_new_array_name(self):
        """Get the name of the new array"""
        return self.__new_name

    def set_take_absolute_value(self, flag):
        """This will take the absolute value of the array before normalization.
        """
        if self.__absolute != flag:
            self.__absolute = flag
            self.Modified()

    def set_normalization(self, norm):
        """Set the normalization operation to perform

        Args:
            norm (str, int, or callable): The operation as a string key, int
                index, or callable method

        Note:
            This can accept a callable method to set a custom operation as long
            as its signature is: ``<callable>(arr)``
        """
        if isinstance(norm, str):
            norm = NormalizeArray.get_normalizations()[norm]
        elif isinstance(norm, int):
            norm = NormalizeArray.get_normalization(norm)
        if self.__normalization != norm:
            self.__normalization = norm
            self.Modified()

    def set_shift(self, sft):
        """Set a static shifter to the input data array"""
        if self.__shift != sft:
            self.__shift = sft
            self.Modified()


###############################################################################


class PercentThreshold(FilterBase):
    """Allows user to select a percent of the data range to threshold.
    This will find the data range of the selected input array and remove the
    bottom percent. This can be reversed using the invert property.
    """
    __displayname__ = 'Percent Threshold'
    __category__ = 'filter'

    def __init__(self, percent=50, invert=False, **kwargs):
        FilterBase.__init__(self, inputType='vtkDataSet',
                            outputType='vtkUnstructuredGrid', **kwargs)
        self.__invert = invert
        if percent < 1.0:
            percent *= 100
        self.__percent = percent # NOTE: not decimal percent
        self.__filter = vtk.vtkThreshold()
        self.__input_array = [None, None]


    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline for execution"""
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        self.__filter.SetInputDataObject(pdi)
        pdo = self.GetOutputData(outInfo, 0)
        # Get Input Array
        field, name = self.__input_array[0], self.__input_array[1]
        wpdi = dsa.WrapDataObject(pdi)
        arr = _helpers.get_numpy_array(wpdi, field, name)

        dmin, dmax = np.nanmin(arr), np.nanmax(arr)
        val = dmin + (self.__percent / 100.0) * (dmax - dmin)

        if self.__invert:
            self.__filter.ThresholdByLower(val)
        else:
            self.__filter.ThresholdByUpper(val)

        self.__filter.Update()

        filt = self.__filter.GetOutputDataObject(0)

        pdo.ShallowCopy(filt)
        return 1


    def SetInputArrayToProcess(self, idx, port, connection, field, name):
        """Used to set the input array(s)

        Args:
            idx (int): the index of the array to process
            port (int): input port (use 0 if unsure)
            connection (int): the connection on the port (use 0 if unsure)
            field (int): the array field (0 for points, 1 for cells, 2 for
                field, and 6 for row)
            name (int): the name of the array
        """
        if self.__input_array[0] != field or self.__input_array[1] != name:
            self.__input_array[0] = field
            self.__input_array[1] = name
            self.__filter.SetInputArrayToProcess(idx, port, connection, field, name)
            self.Modified()
        return 1

    def set_percent(self, percent):
        """Set the percent for the threshold in range (0, 100).
        Any values falling beneath the set percent of the total data range
        will be removed."""
        if self.__percent != percent:
            self.__percent = percent
            self.Modified()

    def set_use_continuous_cell_range(self, flag):
        """If this is on (default is off), we will use the continuous
        interval [minimum cell scalar, maxmimum cell scalar] to intersect
        the threshold bound , rather than the set of discrete scalar
        values from the vertices"""
        return self.__filter.SetUseContinuousCellRange(flag)

    def set_invert(self, flag):
        """Use to invert the threshold filter"""
        if self.__invert != flag:
            self.__invert = flag
            self.Modified()


    def apply(self, input_data_object, array_name):
        """Run the algorithm on an input data object, specifying the array"""
        self.SetInputDataObject(input_data_object)
        arr, field = _helpers.search_for_array(input_data_object, array_name)
        self.SetInputArrayToProcess(0, 0, 0, field, array_name)
        self.Update()
        return pv.wrap(self.GetOutput())

###############################################################################

class ArraysToRGBA(FilterPreserveTypeBase):
    """Use arrays from input data object to set an RGBA array. Sets colors and
    transparencies.
    """
    __displayname__ = 'Arrays To RGBA'
    __category__ = 'filter'

    def __init__(self, **kwargs):
        FilterPreserveTypeBase.__init__(self, **kwargs)
        self.__use_trans = False
        self.__r_array = [None, None]
        self.__g_array = [None, None]
        self.__b_array = [None, None]
        self.__a_array = [None, None]
        self.__field = None
        self.__mask = -9999


    def _get_arrays(self, wpdi):
        """Internal helper to fetch RGBA arrays"""
        # Get Red
        fieldr, name = self.__r_array[0], self.__r_array[1]
        r_arr = _helpers.get_numpy_array(wpdi, fieldr, name)
        # Get Green
        fieldg, name = self.__g_array[0], self.__g_array[1]
        g_arr = _helpers.get_numpy_array(wpdi, fieldg, name)
        # Get Blue
        fieldb, name = self.__b_array[0], self.__b_array[1]
        b_arr = _helpers.get_numpy_array(wpdi, fieldb, name)
        # Get Trans
        fielda, name = self.__a_array[0], self.__a_array[1]
        a_arr = _helpers.get_numpy_array(wpdi, fielda, name)
        if fieldr != fieldg != fieldb: # != fielda
            raise _helpers.PVGeoError('Data arrays must be of the same field.')
        self.__field = fieldr
        return r_arr, g_arr, b_arr, a_arr


    def _mask_arrays(self, r_arr, g_arr, b_arr, a_arr):
        """Internal helper to mask RGBA arrays"""
        r_arr = np.ma.masked_where(r_arr == self.__mask, r_arr)
        g_arr = np.ma.masked_where(g_arr == self.__mask, g_arr)
        b_arr = np.ma.masked_where(b_arr == self.__mask, b_arr)
        a_arr = np.ma.masked_where(a_arr == self.__mask, a_arr)
        return r_arr, g_arr, b_arr, a_arr


    def RequestData(self, request, inInfo, outInfo):
        """Execute on pipeline"""
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        wpdi = dsa.WrapDataObject(pdi)
        # Get number of points
        pdo = self.GetOutputData(outInfo, 0)

        # Get the arrays for the RGB values
        r_arr, g_arr, b_arr, a_arr = self._get_arrays(wpdi)
        r_arr, g_arr, b_arr, a_arr = self._mask_arrays(r_arr, g_arr, b_arr, a_arr)

        # normalize each color array bewteen 0 and 255
        r_arr = NormalizeArray._feature_scale(r_arr, [0, 255])
        g_arr = NormalizeArray._feature_scale(g_arr, [0, 255])
        b_arr = NormalizeArray._feature_scale(b_arr, [0, 255])

        # Now concatenate the arrays
        if self.__use_trans:
            a_arr = NormalizeArray._feature_scale(a_arr, [0, 255])
            col = np.array(np.c_[r_arr, g_arr, b_arr, a_arr], dtype=np.uint8)
        else:
            col = np.array(np.c_[r_arr, g_arr, b_arr], dtype=np.uint8)
        colors = interface.convert_array(col, name='Colors')

        # Set the output
        pdo.DeepCopy(pdi)
        # Add new color array
        _helpers.add_array(pdo, self.__field, colors)
        return 1



    #### Seters and Geters ####


    def set_use_transparency(self, flag):
        """Set a boolean flag on whether or not to use a transparency component"""
        if self.__use_trans != flag:
            self.__use_trans = flag
            self.Modified()

    def set_mask_value(self, val):
        """Set the value to mask in the RGBA arrays"""
        if self.__mask != val:
            self.__mask = val
            self.Modified()

    def _set_input_array_red(self, field, name):
        """Set field and name of red array"""
        if self.__r_array[0] != field:
            self.__r_array[0] = field
            self.Modified()
        if self.__r_array[1] != name:
            self.__r_array[1] = name
            self.Modified()

    def _set_input_array_green(self, field, name):
        """Set field and name of green array"""
        if self.__g_array[0] != field:
            self.__g_array[0] = field
            self.Modified()
        if self.__g_array[1] != name:
            self.__g_array[1] = name
            self.Modified()

    def _set_input_array_blue(self, field, name):
        """Set field and name of blue array"""
        if self.__b_array[0] != field:
            self.__b_array[0] = field
            self.Modified()
        if self.__b_array[1] != name:
            self.__b_array[1] = name
            self.Modified()

    def _set_input_array_trans(self, field, name):
        """Set field and name of transparency array"""
        if self.__a_array[0] != field:
            self.__a_array[0] = field
            self.Modified()
        if self.__a_array[1] != name:
            self.__a_array[1] = name
            self.Modified()

    def SetInputArrayToProcess(self, idx, port, connection, field, name):
        """Used to set the input array(s)

        Args:
            idx (int): the index of the array to process
            port (int): input port (use 0 if unsure)
            connection (int): the connection on the port (use 0 if unsure)
            field (int): the array field (0 for points, 1 for cells, 2 for
                field, and 6 for row)
            name (int): the name of the array
        """
        if idx == 0:
            self._set_input_array_red(field, name)
        elif idx == 1:
            self._set_input_array_green(field, name)
        elif idx == 2:
            self._set_input_array_blue(field, name)
        elif idx == 3:
            self._set_input_array_trans(field, name)
        else:
            raise _helpers.PVGeoError('SetInputArrayToProcess() do not know how to handle idx: %d' % idx)
        return 1

    def apply(self, input_data_object, r_array, g_array, b_array, a_array=None):
        """Run the algorithm on an input data object, specifying RGBA array names"""
        self.SetInputDataObject(input_data_object)
        r_arr, rField = _helpers.search_for_array(input_data_object, r_array)
        g_arr, gField = _helpers.search_for_array(input_data_object, g_array)
        b_arr, bField = _helpers.search_for_array(input_data_object, b_array)
        if a_array is not None:
            a_arr, aField = _helpers.search_for_array(input_data_object, a_array)
            self.SetInputArrayToProcess(3, 0, 0, aField, a_array)
            self.set_use_transparency(True)
        self.SetInputArrayToProcess(0, 0, 0, rField, r_array)
        self.SetInputArrayToProcess(1, 0, 0, gField, g_array)
        self.SetInputArrayToProcess(2, 0, 0, bField, b_array)
        self.Update()
        return pv.wrap(self.GetOutput())



###############################################################################



"""
Export
======

On this page, we demonstrate how to quickly share a 3D rendering of your
ParaView visualizations with anyone who has access to the internet so that that
can explore the whole scene in a dynamic manner.

Motivation
----------
In order to effectively communicate our geoscientific findings, we often need to
share our 3D visualizations with interested stakeholders. These interested
parties are likely not going to have ParaView or other visualization software
at hand. Thus we desire to have a means to export our complex visualizations in
ParaView to a simple, shareable format that anyone can view. To accomplish this,
we will take advantage of vtk.js and its standalone web viewer for vtk.js
formats.

Would not it be great to send your client or interested parties an interactive
3D scene of your Geophysical findings like the example below?

.. raw:: html

    <div style="position: relative; padding-bottom: 56.25%; height: 0; overflow: hidden; max-width: 100%; height: auto;">
            <iframe src="http://tunnels.pvgeo.org/" frameborder="0" allowfullscreen style="position: absolute; top: 0; left: 0; width: 100%; height: 100%;"></iframe>
    </div>



VTK.js
------

`vtk.js <https://kitware.github.io/vtk-js/>`_ is a rendering library made for
scientific visualization on the web. This code base brings high performance
rendering into anyone's web browser. This library allows us to export complex
scenes from ParaView and share them with anyone that has a web browser like
Safari or Google Chrome.

The vtk.js library has an open-source `standalone scene viewer <https://kitware.github.io/vtk-js/examples/StandaloneSceneLoader/StandaloneSceneLoader.html>`_
which they have a `nice demo <https://kitware.github.io/vtk-js/examples/StandaloneSceneLoader.html>`_.
The first link can either be downloaded as an HTML file to be ran locally, or
you can go to that link and run from the vtk.js server. vtk.js also published a
scene export macro for ParaView that compresses a data scene in ParaView to a
shareable format for viewing on the web. The `macro from the vtk.js library <https://raw.githubusercontent.com/Kitware/vtk-js/master/Utilities/ParaView/export-scene-macro.py>`_
can be used but we also deploy an updated (we think more robust) version of
this export macro in the sub-module `export` of our Python module `pvmacros`.

Demo Shareable Format
+++++++++++++++++++++

Here are some samples to demonstrate the web viewer which we host on
`viewer.pyvista.org <http://viewer.pyvista.org>`_. We have included a few of our
scenes and one of the vtk.js sample scenes for you to demo:

- `Fluvial Channels <http://viewer.pyvista.org/?fileURL=https://dl.dropbox.com/s/qnahdwedjwndo7t/fluvsim_channels.vtkjs?dl=0>`_
- `Volcano <http://volcano.pvgeo.org>`_
- `Ripple <http://ripple.pvgeo.org>`_
- `Tunnels <http://tunnels.pvgeo.org>`_
- `vtk.js Sample Scene <http://viewer.pyvista.org/?fileURL=https://data.kitware.com/api/v1/file/587003c38d777f05f44a5c93/download>`_



Example Use
-----------

First, make a complex scene in ParaView that you might like to share with someone.
Now that you have your scene loaded, open the python shell from'View->Python Shell'
(or 'Tools->Python Shell' depending on your ParaView version) within ParaView.
From here, import our Python module delivered in the repository called
``#!py pvmacros``. From the ``#!py export`` sub-module, there is a function called
``#!py def exportVTKjs()`` which takes two optional arguments (`FileName` string
and `compress` boolean). Execute this function and note the output text as it
will describe where the exported scene was saved.

.. code-block:: python

    ## Import our ParaView Macros module:
    import pvmacros as pvm

    ## Now run the exportVTKjs script from the export sub-module
    pvm.export.exportVTKjs(FileName='test_export')


Now open the standalone web viewer by opening `viewer.pyvista.org <http://viewer.pyvista.org>`_

Select the exported scene as the input file for the web viewer from where you
saved it (should be under ``~/Dropbox/PVGeo_vtkjs/``). The export macro should
have printed out the location of the saved scene in the Python Shell.

If you have trouble post on our `issues page <https://github.com/OpenGeoVis/PVGeo/issues>`_
or read the `vtk.js documentation <https://kitware.github.io/vtk-js/examples/StandaloneSceneLoader.html>`_


How to Share
------------

To share these exported scenes with non-technical stakeholders, we recommend the
following processes:

Quick and Easy
++++++++++++++

- Create your scene and export to the vtk.js format (follow process above)
- Quality control your visualization by viewing in web browser yourself (follow process above)
- Send an email with your visualization (``*.vtkjs`` file) and something along the lines of:

.. code-block:: text

    Check out the data scene/model by downloading the attached file.
    Then go to the link below and open that downloaded file.

    `http://viewer.pyvista.org/ <http://viewer.pyvista.org>`_


A Bit More Robust
+++++++++++++++++

Sometimes we might want to give someone a direct link to the web visualization
so all they have to do is open the link on any device and they can see our
visualization. Here is a method to share scenes that have a slightly easier
process of viewing the file for the end user and will handle the case for mobile
platforms.

Unfortunately, making the experience for the end user simple means making your
experience a bit more complicated. You will need to host your file on a web
service like GitHub or Dropbox *(we have been unsuccessful in getting Google
Drive to work)*. Then get a public link to the `*.vtkjs` file on that web
service and append it to the web viewer URL.

We have created a Python script to generate these links for you if you are
sharing your data file on either Dropbox or GitHub. The script is delivered
in the repository.

The easiest way that we have found is to share the files on Dropbox.
Use the desktop client for Dropbox and right-click your exported `*.vtkjs` file
and select "Copy Dropbox Link."

Once you have that link, use the this script on your URLs in this manner:


.. code-block:: bash

    ## Usage:
    python get_vtkjs_url.py <web file host> <file link>

    ## Dropbox example:
    python get_vtkjs_url.py dropbox "https://www.dropbox.com/s/6m5ttdbv5bf4ngj/ripple.vtkjs?dl=0"



.. raw:: html

    <iframe src="https://player.vimeo.com/video/257833915" width="640" height="360" frameborder="0" webkitallowfullscreen mozallowfullscreen allowfullscreen></iframe>
    <p><a href="https://vimeo.com/257833915">PVGeo Export Demo</a> from <a href="https://vimeo.com/user82050125">Bane Sullivan</a> on <a href="https://vimeo.com">Vimeo</a>.</p>

"""

from .vtkjs import *

__displayname__ = 'Export'
